from collections import namedtuple

from .General import dimensions_to_size, get_NUM_token_value
from .Token import Token

"""
This file holds the program's functions and global variables
(as global variables, hehe)
And related functions
"""

global_variables = list()  # Global list of global variables


# variables
def get_global_variables():
    return global_variables


def insert_global_variable(variable):
    get_global_variables().append(variable)


def get_global_variables_size():
    return sum(get_variable_size(variable) for variable in get_global_variables())


def create_variable(name, type, dimensions):
    # return variable named tuple
    variable = namedtuple("variable", ["name", "type", "size", "cell_index"])

    variable.name = name
    variable.type = type
    variable.dimensions = dimensions  # list of array dimensions sizes (for non-arrays it will be [1])
    variable.cell_index = None  # will be updated when we insert this variable into an ids map

    return variable


def get_variable_size(variable):
    # return total variable size
    return dimensions_to_size(variable.dimensions)


def get_variable_dimensions(variable):
    return variable.dimensions


def is_variable_array(variable):
    return variable.dimensions != [1]


def create_variable_from_definition(parser, index=None, advance_tokens=False):
    """
    processes the variable definition at index, and returns the variable named tuple
    if index is None, then assumes we start at the current_token_index
    if advance_tokens is True, then modifies current_token_index accordingly using parser.advance_token()
    """

    if index is None:
        index = parser.current_token_index

    assert parser.tokens[index].type == Token.INT

    parser.check_next_token_is(Token.ID, starting_index=index)
    ID = parser.tokens[index + 1].data

    if advance_tokens:
        parser.advance_token(amount=2)  # skip INT ID

    if parser.tokens[index + 2].type == Token.LBRACK:  # array (support multi-dimensional arrays)
        dimensions = []  # element[i] holds the size of dimension[i]
        while parser.tokens[index + 2].type == Token.LBRACK:
            parser.check_next_tokens_are([Token.LBRACK, Token.NUM, Token.RBRACK], starting_index=index + 1)
            dimensions.append(get_NUM_token_value(parser.tokens[index + 3]))

            if advance_tokens:
                parser.advance_token(amount=3)  # skip LBRACK NUM RBRACK
            index += 3
    else:
        dimensions = [1]

    return create_variable(ID, Token.INT, dimensions)



"""Perform average / max pooling on a matrix"""

from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import numpy as np

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Your job is to perform max/average pooling on the given matrix.
The stride is equal to the kernel size, meaning there is no overlap between the pooling regions.

Your output should be a matrix in the same format as the input matrix.
The output matrix is smaller than the input matrix when the kernel size is greater than 1, and its elements may be floating-point numbers.
Give elements in the output matrix correct to 2 decimal places.

Perform {pool_type} pooling on the following matrix with a kernel size of {pool_size}:
{matrix}
"""


DATASET_NAME = "pool_matrix"


@dataclass
class PoolMatrixConfig:
    """Configuration for Pool Matrix dataset generation"""

    min_rows: int = 2  # Minimum rows of the matrix
    max_rows: int = 10  # Maximum rows of the matrix
    min_cols: int = 2  # Minimum columns of the matrix
    max_cols: int = 10  # Maximum columns of the matrix
    min_pool_size: int = 1  # Minimum pooling size
    max_pool_size: int = 3  # Maximum pooling size

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 2 <= self.min_rows, "min_rows must be at least 2"
        assert self.min_rows <= self.max_rows, "max_rows must be at least min_rows"
        assert 2 <= self.min_cols, "min_cols must be at least 2"
        assert self.min_cols <= self.max_cols, "max_cols must be at least min_cols"
        assert 1 <= self.min_pool_size, "min_pool_size must be at least 1"
        assert self.min_pool_size <= self.max_pool_size, "max_pool_size must be at least min_pool_size"


class PoolMatrixDataset(ProceduralDataset):
    """Generates Pool Matrix exercises with configurable difficulty"""

    def __init__(self, config: PoolMatrixConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _matrix_to_str(self, matrix: np.ndarray) -> str:
        """Get a string representation of the matrix"""
        return "\n".join(" ".join(str(round(x, 2)) for x in row) for row in matrix)

    def _max_pool(self, matrix: np.ndarray, pool_size: int) -> np.ndarray:
        """Perform max pooling on the matrix"""
        rows, cols = matrix.shape
        return np.array(
            [
                [np.max(matrix[i : i + pool_size, j : j + pool_size]) for j in range(0, cols, pool_size)]
                for i in range(0, rows, pool_size)
            ]
        )

    def _average_pool(self, matrix: np.ndarray, pool_size: int) -> np.ndarray:
        """Perform average pooling on the matrix"""
        rows, cols = matrix.shape
        return np.array(
            [
                [np.mean(matrix[i : i + pool_size, j : j + pool_size]) for j in range(0, cols, pool_size)]
                for i in range(0, rows, pool_size)
            ]
        )

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Score the answer based on the metadata"""

        if not isinstance(answer, str):
            return 0.0

        reward = 0.0
        try:
            oracle_answer = np.loadtxt(entry["answer"].splitlines(), dtype=np.float32)
            answer = np.loadtxt(answer.splitlines(), dtype=np.float32)
            if oracle_answer.shape == answer.shape and np.allclose(oracle_answer, answer, rtol=1e-2):
                reward = 1.0
            elif oracle_answer.shape == answer.shape:
                reward = 0.1
        except Exception:
            pass
        return reward

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Pool Matrix question"""
        rng = Random(self.seed + idx)
        np.random.seed(self.seed + idx)

        rows = rng.randint(self.config.min_rows, self.config.max_rows)
        cols = rng.randint(self.config.min_rows, self.config.max_cols)
        matrix = np.random.randint(0, 10, (rows, cols))
        matrix_str = self._matrix_to_str(matrix)

        pool_size = rng.randint(self.config.min_pool_size, self.config.max_pool_size)
        pool_type = rng.choice(["average", "max"])

        answer = self._average_pool(matrix, pool_size) if pool_type == "average" else self._max_pool(matrix, pool_size)
        answer_str = self._matrix_to_str(answer)

        return {
            "question": QUESTION_TEMPLATE.format(matrix=matrix_str, pool_type=pool_type, pool_size=pool_size),
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "matrix": matrix.tolist(),
                "pool_type": pool_type,
                "pool_size": pool_size,
                "solution": answer.tolist(),
                "rows": rows,
                "cols": cols,
                "pool_size": pool_size,
                "difficulty": {
                    "rows": (self.config.min_rows, self.config.max_rows),
                    "cols": (self.config.min_cols, self.config.max_cols),
                    "pool_size": (self.config.min_pool_size, self.config.max_pool_size),
                },
            },
        }


class PoolMatrixCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(PoolMatrixCurriculum.__name__, PoolMatrixConfig)

        self._define_attributes(
            RangeAttributeDefinition(
                name="rows",
                levels=[10, 25, 50, 100],
                description="Board size",
                lower_field_name="min_rows",
                upper_field_name="max_rows",
            ),
            RangeAttributeDefinition(
                name="cols",
                levels=[10, 25, 50, 100],
                description="Board size",
                lower_field_name="min_cols",
                upper_field_name="max_cols",
            ),
            RangeAttributeDefinition(
                name="pool_size",
                levels=[3, 5, 7, 9],
                description="Pool size",
                lower_field_name="min_pool_size",
                upper_field_name="max_pool_size",
            ),
        )


register_dataset(DATASET_NAME, PoolMatrixDataset, PoolMatrixConfig, PoolMatrixCurriculum)



"""
Survo dataset, adapted for Reasoning Gym from SynthRL: https://github.com/MiniMax-AI/SynLogic/tree/main/games/tasks/survo
"""

from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import numpy as np

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "survo"

PROMPT_TEMPLATES = [
    "Given a {n}*{n} matrix where the last element of each row and column equals the sum of the other elements in that row or column. The matrix is:\n{matrix}\nwhere some elements are replaced with 0. You have a set of numbers {numbers} that can be filled into the 0 positions to satisfy the rules. Please fill in the matrix. Each number can only be used once.",
    "You have a {n}*{n} matrix with some positions already filled with numbers and others marked with 0. The matrix is:\n{matrix}\nThe last number in each row and column represents the sum of all other numbers in that row or column. You need to fill in the 0 positions using the numbers {numbers} to satisfy these conditions. Each number can only be used once.",
    "Complete the following Survo puzzle. In this {n}*{n} matrix:\n{matrix}\nthe cells marked with 0 need to be filled with numbers. The last number in each row and column equals the sum of all other numbers in that row or column. You can use the following numbers: {numbers}. Each number can only be used once.",
    "In this {n}*{n} Survo matrix puzzle:\n{matrix}\nthe 0 cells need to be filled with numbers from the set {numbers}. The last element in each row and column is the sum of all other elements in that row or column. Each number can only be used once. Provide the completed matrix.",
    "Solve this {n}*{n} matrix puzzle:\n{matrix}\nwhere 0 represents empty cells that need to be filled. The last number in each row and column equals the sum of all other numbers in that row or column. You have the numbers {numbers} to place in the empty cells. Each number can only be used once.",
]


@dataclass
class SurvoConfig:
    min_board_size: int = 4
    max_board_size: int = 5
    min_empty: int = 3
    max_empty: int = 5
    min_num: int = 1
    max_num: int = 9
    seed: Optional[int] = None
    size: int = 500  # Virtual dataset size

    def validate(self):
        """Validate configuration parameters"""
        assert self.min_board_size > 3, "min_board_size must be greater than 3"
        assert self.max_board_size >= self.min_board_size, "max_board_size must be >= min_board_size"
        assert self.min_empty > 0, "min_empty must be > 0"
        assert self.max_empty <= (self.min_board_size - 1) * (
            self.min_board_size - 1
        ), f"max_empty must be <= {(self.min_board_size - 1) * (self.min_board_size - 1)}"
        assert self.min_empty <= self.max_empty, "min_empty must be <= max_empty"
        assert self.min_num > 0, "min_num must be > 0"
        assert self.min_num < self.max_num, "min_num must be less than max_num"


class SurvoDataset(ProceduralDataset):
    def __init__(self, config: SurvoConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __len__(self) -> int:
        return self.config.size

    def __iter__(self):
        self._current_idx = 0
        return self

    def __next__(self):
        if self._current_idx >= self.config.size:
            raise StopIteration
        item = self[self._current_idx]
        self._current_idx += 1
        return item

    def __getitem__(self, idx: int) -> dict:
        rng = Random(self.config.seed + idx)

        board_size = rng.randint(self.config.min_board_size, self.config.max_board_size)
        num_empty = rng.randint(self.config.min_empty, self.config.max_empty)

        filled_matrix, puzzle, candidate_numbers = self._generate_valid_matrix(
            rng, board_size, num_empty, self.config.min_num, self.config.max_num
        )

        puzzle_str = "\n".join(" ".join(str(x) for x in row) for row in puzzle)
        solution_str = "\n".join(" ".join(str(x) for x in row) for row in filled_matrix)

        question = rng.choice(PROMPT_TEMPLATES).format(n=board_size, matrix=puzzle_str, numbers=candidate_numbers)

        return {
            "question": question,
            "answer": solution_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_idx": idx,
                "puzzle": puzzle.tolist(),
                "solution": filled_matrix.tolist(),
                "candidate_numbers": candidate_numbers,
                "board_size": board_size,
                "num_empty": num_empty,
                "min_num": self.config.min_num,
                "max_num": self.config.max_num,
                "difficulty": {
                    "board_size": (self.config.min_board_size, self.config.max_board_size),
                    "empty": (self.config.min_empty, self.config.max_empty),
                },
            },
        }

    def _generate_valid_matrix(
        self, rng: Random, n: int, num_empty: int, min_num: int, max_num: int
    ) -> tuple[np.ndarray, np.ndarray, list[int]]:
        matrix = np.zeros((n, n), dtype=int)

        for i in range(n - 1):
            for j in range(n - 1):
                matrix[i, j] = rng.randint(min_num, max_num)

        for i in range(n - 1):
            row_sum = sum(matrix[i, 0 : n - 1])
            matrix[i, n - 1] = row_sum

            col_sum = sum(matrix[0 : n - 1, i])
            matrix[n - 1, i] = col_sum

        matrix[n - 1, n - 1] = sum(matrix[0 : n - 1, n - 1])

        filled_matrix = matrix.copy()

        positions = [(i, j) for i in range(n - 1) for j in range(n - 1)]
        selected_positions = rng.sample(positions, num_empty)

        candidate_numbers = []
        for i, j in selected_positions:
            candidate_numbers.append(int(matrix[i, j]))
            matrix[i, j] = 0

        return filled_matrix, matrix, candidate_numbers

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        if not isinstance(answer, str):
            return 0.0

        board_size = entry["metadata"]["board_size"]
        grid = self._parse_grid(answer)
        true_grid = entry["metadata"]["solution"]

        if len(grid) != board_size or any(len(row) != board_size for row in grid):
            return 0.0

        for i in range(board_size):
            for j in range(board_size):
                if grid[i][j] != true_grid[i][j]:
                    return 0.0

        return 1.0

    def _parse_grid(self, answer: str) -> list[list[str]]:
        grid = []
        for line in answer.strip().split("\n"):
            row = []
            for c in line.strip().split():
                try:
                    row.append(int(c))
                except ValueError:
                    continue  # Ignore non-integer values
            if row:
                grid.append(row)
        return grid


class SurvoCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(SurvoCurriculum.__name__, SurvoConfig)

        self._define_attributes(
            RangeAttributeDefinition(
                name="board_size",
                levels=[4, 5, 6, 7],
                description="Board size (n x n)",
                lower_field_name="min_board_size",
                upper_field_name="max_board_size",
            ),
            RangeAttributeDefinition(
                name="empty",
                levels=[4, 9, 16, 25],
                description="Number of empty cells",
                lower_field_name="min_empty",
                upper_field_name="max_empty",
            ),
        )


register_dataset(DATASET_NAME, SurvoDataset, SurvoConfig, SurvoCurriculum)



# -*- coding: utf-8 -*-


import json
from typing import Any, Dict, List

from reasoning_gym.data import get_data_file_path

from .const import (
    ALL_CONFIG_SIZE,
    ONOFFOFF_MAX_NON,
    ONOFFOFF_MAX_POTENTIAL,
    ONOFFOFF_MIN_NON,
    ONOFFOFF_MIN_POTENTIAL,
    ONONOFF_MAX_NON,
    ONONOFF_MAX_POTENTIAL,
    ONONOFF_MIN_NON,
    ONONOFF_MIN_POTENTIAL,
)


class BlicketView(object):
    def __init__(self, light_state="no"):
        # light state: "no" for no light, "off" for light off, "on" for light on
        self.objects = []
        self.light_state = light_state

    def add_objects(self, objects):
        for obj in objects:
            self.objects.append(obj)

    def remove_objects(self, objects):
        for obj in objects:
            self.objects.remove(obj)

    def __repr__(self):
        return "BlicketView(objects={}, light_state={})".format(self.objects, self.light_state)


class BlicketQuestion(object):
    def __init__(
        self,
        min_potential_blickets,
        max_potential_blickets,
        min_non_blickets,
        max_non_blickets,
        config_size,
        shuffle,
        rng,
    ):
        self.min_potential_blickets = min_potential_blickets
        self.max_potential_blickets = max_potential_blickets
        self.min_non_blickets = min_non_blickets
        self.max_non_blickets = max_non_blickets
        self.config_size = config_size
        self.shuffle = shuffle
        self.rng = rng

        potential_blicket_num = self.rng.randint(self.min_potential_blickets, self.max_potential_blickets)
        non_blicket_num = self.rng.randint(self.min_non_blickets, self.max_non_blickets)

        samples = self.rng.sample(list(range(self.config_size)), k=potential_blicket_num + non_blicket_num)

        self.blickets = []
        self.set_blickets = []
        self.potential_blickets = samples[:potential_blicket_num]
        self.non_blickets = samples[potential_blicket_num:]

        self.direct = []
        self.indirect = []
        self.screen_off = []

    def get_habituation_views(self):
        blicket_obj = self.potential_blickets[0]
        self.add_blicket(blicket_obj)

        non_blicket_obj = self.non_blickets[0]

        view_with_blicket = BlicketView("on")
        view_with_blicket.add_objects([blicket_obj])

        view_with_non_blicket = BlicketView("off")
        view_with_non_blicket.add_objects([non_blicket_obj])

        view_with_both = BlicketView("on")
        view_with_both.add_objects([blicket_obj, non_blicket_obj])

        habituation_views = [view_with_blicket, view_with_non_blicket, view_with_both]

        # bookkeeping for candidate choices
        self.direct.append(blicket_obj)
        self.screen_off.append(non_blicket_obj)

        return habituation_views

    def get_evidence_views(self):
        raise NotImplementedError("The parent class should not be used.")

    def get_views(self):
        habituation_views = self.get_habituation_views()
        evidence_views = self.get_evidence_views()
        if self.shuffle:
            self.rng.shuffle(habituation_views)
            self.rng.shuffle(evidence_views)

        return habituation_views + evidence_views

    def sanity_check(self):
        # no duplicates
        assert len(self.direct) == len(set(self.direct))
        assert len(self.indirect) == len(set(self.indirect))
        assert len(self.screen_off) == len(set(self.screen_off))
        # no intersection
        assert len(set(self.direct).intersection(self.indirect)) == 0
        assert len(set(self.direct).intersection(self.screen_off)) == 0
        assert len(set(self.indirect).intersection(self.screen_off)) == 0
        # completeness
        assert set(self.direct + self.indirect + self.screen_off) == set(self.blickets + self.non_blickets)

    def check_blickets(self, views):
        blickets = set()
        non_blickets = set()
        potential_blickets = set()

        direct = set()
        indirect = set()
        screen_off = set()

        on_views = [view for view in views if view.light_state == "on"]
        off_views = [view for view in views if view.light_state == "off"]

        assert len(on_views) + len(off_views) == len(views)

        for off_view in off_views:
            non_blickets.update(off_view.objects)
        for on_view in on_views:
            if len(on_view.objects) == 1:
                blickets.update(on_view.objects)
            else:
                diff_set = set(on_view.objects).difference(non_blickets)
                if len(diff_set) == 1:
                    blickets.update(diff_set)
        all_objects = set()
        for view in views:
            all_objects.update(view.objects)
        potential_blickets.update(all_objects.difference(non_blickets).difference(blickets))

        assert blickets == set(self.blickets)
        assert non_blickets == set(self.non_blickets)
        assert potential_blickets == set(self.potential_blickets)

        for on_view in on_views:
            if len(on_view.objects) == 1:
                direct.update(on_view.objects)
        on_view_objects = set()
        for on_view in on_views:
            on_view_objects.update(on_view.objects)
        off_view_objects = set()
        for off_view in off_views:
            off_view_objects.update(off_view.objects)

        direct.update(off_view_objects.difference(on_view_objects))
        screen_off.update(off_view_objects.intersection(on_view_objects))

        for on_view in on_views:
            if len(on_view.objects) > 1:
                diff_set = set(on_view.objects).difference(non_blickets)
                if len(diff_set) == 1 and not diff_set.issubset(direct):
                    indirect.update(diff_set)

        assert direct == set(self.direct)
        assert indirect == set(self.indirect)
        assert screen_off == set(self.screen_off)

        set_blickets = set()
        for on_view in on_views:
            on_diff_set = set(on_view.objects).difference(non_blickets)
            if len(on_diff_set.intersection(blickets)) == 0:
                potential_set_blicket = list(on_diff_set)
                potential_set_blicket.sort()
                set_blickets.add(tuple(potential_set_blicket))
        remove_set = set()
        for elem in set_blickets:
            for another_elem in set_blickets:
                elem_set = set(elem)
                another_elem_set = set(another_elem)
                if elem_set < another_elem_set:
                    remove_set.add(another_elem)
        set_blickets.difference_update(remove_set)

        assert set_blickets == set(self.set_blickets), "set_blickets:{}, self.set_blickets:{}".format(
            set_blickets, self.set_blickets
        )

    def check_labels(self, view, label):
        diff_set = set(view.objects).difference(self.non_blickets)
        if len(diff_set) == 0:
            assert label == 0
        else:
            blicket_inter = diff_set.intersection(self.blickets)
            if len(blicket_inter) > 0:
                assert label == 2
            else:
                if self.has_set_blicket(view):
                    assert label == 2
                else:
                    assert label == 1

    def has_set_blicket(self, view):
        for set_blicket in self.set_blickets:
            if set(set_blicket).issubset(view.objects):
                return True
        return False

    def union_sample(self, union, must_have_one=False):
        if must_have_one:
            first_set = self.rng.sample(union, k=1)
        else:
            first_num = self.rng.randint(1, len(union))
            first_set = self.rng.sample(union, k=first_num)
        second_set = list(set(union).difference(first_set))
        if len(second_set) > 0:
            additional_num_min = 0
        else:
            additional_num_min = 1
        additional_num = self.rng.randint(additional_num_min, len(first_set))
        additional_samples = self.rng.sample(first_set, k=additional_num)
        second_set += additional_samples

        return first_set, second_set

    def add_noise(self, view):
        # the first non blicket used in habituation and hence the skip
        noise_num = self.rng.randint(0, len(self.non_blickets) - 1)
        noise = self.rng.sample(self.non_blickets[1:], k=noise_num)
        view.add_objects(noise)

    def add_blicket(self, obj):
        self.potential_blickets.remove(obj)
        self.blickets.append(obj)

    def add_set_blicket(self, set_blicket):
        to_remove_set = []
        for already_set_blicket in self.set_blickets:
            if set(set_blicket) < set(already_set_blicket):
                to_remove_set.append(already_set_blicket)
            if set(already_set_blicket) < set(set_blicket):
                to_remove_set.append(set_blicket)
        self.set_blickets.append(set_blicket)
        for elem in to_remove_set:
            self.set_blickets.remove(elem)

    def generate_cause_questions(self, train, regime="IID"):

        def fixed_sum_sample(lower, upper, fixed_sum):
            values = [0] * len(lower)
            assert sum(upper) >= fixed_sum
            while True:
                residual = fixed_sum
                for i in range(len(values) - 1):
                    values[i] = self.rng.randint(lower[i], min(upper[i], residual))
                    residual = residual - values[i]
                if residual >= lower[-1] and residual <= upper[-1]:
                    values[-1] = residual
                    break
            assert sum(values) == fixed_sum
            return values

        # direct_sample, indirect_sample, screen_off_sample, potential_sample
        # these magic numbers and changes are to adjust dataset statistics
        constraint_lower = [0] * 3
        if regime == "Comp":
            # for Comp split
            constraint_upper = [len(self.direct), len(self.indirect), max(len(self.screen_off) - 2, 0)]
        if regime == "Sys":
            if train:
                # for Sys train split
                constraint_upper = [len(self.direct), len(self.indirect), max(len(self.screen_off) - 3, 0)]
            else:
                # for Sys val / test split
                constraint_upper = [len(self.direct), len(self.indirect), max(len(self.screen_off) - 1, 0)]
        if regime == "IID":
            # for IID split
            constraint_upper = [len(self.direct), len(self.indirect), max(len(self.screen_off) - 1, 0)]
        potential_sample_num = min(len(self.potential_blickets), 1)
        direct_sample_num, indirect_sample_num, screen_off_sample_num = fixed_sum_sample(
            constraint_lower, constraint_upper, 2 - potential_sample_num
        )
        direct_samples = self.rng.sample(self.direct, k=direct_sample_num)
        indirect_samples = self.rng.sample(self.indirect, k=indirect_sample_num)
        screen_off_samples = self.rng.sample(self.screen_off, k=screen_off_sample_num)
        potential_samples = self.rng.sample(self.potential_blickets, k=potential_sample_num)

        questions = []
        # label: 0 for light off, 1 for unknown, 2 for light up
        for direct_sample in direct_samples:
            assert direct_sample in self.blickets or direct_sample in self.non_blickets
            if direct_sample in self.blickets:
                label = 2
            else:
                label = 0
            cause_view = BlicketView("no")
            cause_view.add_objects([direct_sample])
            questions.append((cause_view, label, "direct"))
        for indirect_sample in indirect_samples:
            assert indirect_sample in self.blickets
            label = 2
            cause_view = BlicketView("no")
            cause_view.add_objects([indirect_sample])
            questions.append((cause_view, label, "indirect"))
        for screen_off_sample in screen_off_samples:
            assert screen_off_sample in self.non_blickets
            label = 0
            cause_view = BlicketView("no")
            cause_view.add_objects([screen_off_sample])
            questions.append((cause_view, label, "screen_off"))
        for potential_sample in potential_samples:
            assert potential_sample in self.potential_blickets
            label = 1
            cause_view = BlicketView("no")
            cause_view.add_objects([potential_sample])
            questions.append((cause_view, label, "potential"))

        if self.shuffle:
            self.rng.shuffle(questions)

        return questions

    def generate_intervention_questions(self, views):

        # on_views = [view for view in views if view.light_state == "on" and len(view.objects) >= 2]
        off_views = [view for view in views if view.light_state == "off"]

        # on_view = rng.sample(on_views, k=1)[0]
        # on_view_ref = views.index(on_view)
        off_view = self.rng.sample(off_views, k=1)[0]
        off_view_ref = views.index(off_view)

        questions = []

        all_possibilities = (
            list(set(self.blickets + self.potential_blickets + self.non_blickets).difference(off_view.objects))
            + self.set_blickets
        )
        # adjust weight during sample for better statistics
        all_possibilities += list(set(self.potential_blickets).difference(off_view.objects))
        possibilities = self.rng.sample(all_possibilities, k=2)
        for possibility in possibilities:
            if possibility in self.direct:
                q_type = "direct"
            elif possibility in self.indirect:
                q_type = "indirect"
            elif possibility in self.screen_off:
                q_type = "screen_off"
            elif possibility in self.potential_blickets:
                q_type = "potential"
            else:
                assert possibility in self.set_blickets
                q_type = "indirect"
            if possibility in self.blickets:
                label = 2
            elif possibility in self.potential_blickets:
                label = 1
            elif possibility in self.non_blickets:
                label = 0
            else:
                assert possibility in self.set_blickets
                label = 2
            intervention_view = BlicketView("no")
            intervention_view.add_objects(off_view.objects)
            if type(possibility) == tuple:
                possibility = list(possibility)
            else:
                possibility = [possibility]
            intervention_view.add_objects(possibility)
            questions.append((intervention_view, label, q_type, off_view_ref))

        if self.shuffle:
            self.rng.shuffle(questions)

        return questions


class OnOffOff(BlicketQuestion):
    def __init__(
        self,
        min_potential_blickets,
        max_potential_blickets,
        min_non_blickets,
        max_non_blickets,
        config_size,
        shuffle,
        rng,
    ):
        super(OnOffOff, self).__init__(
            min_potential_blickets,
            max_potential_blickets,
            min_non_blickets,
            max_non_blickets,
            config_size,
            shuffle,
            rng,
        )

    def get_evidence_views(self):
        # the first non blicket used in habituation and hence the skip
        first_off, second_off = self.union_sample(self.non_blickets[1:])

        first_off_view = BlicketView("off")
        first_off_view.add_objects(first_off)

        second_off_view = BlicketView("off")
        second_off_view.add_objects(second_off)

        on_view = BlicketView("on")

        on_list = self.potential_blickets[:]

        if len(on_list) == 1:
            self.add_blicket(on_list[0])
        else:
            set_blicket = on_list[:]
            set_blicket.sort()
            set_blicket = tuple(set_blicket)
            self.add_set_blicket(set_blicket)

        on_view.add_objects(on_list)

        self.add_noise(on_view)

        evidence_views = [on_view, first_off_view, second_off_view]

        # bookkeeping for candidate choices
        off_union_set = set(self.non_blickets[1:])
        on_set = set(on_view.objects)
        on_diff_set = on_set.difference(off_union_set)
        direct_set = list(off_union_set.difference(on_set))
        screen_off_set = list(off_union_set.intersection(on_set))

        self.direct.extend(direct_set)
        self.screen_off.extend(screen_off_set)

        if len(on_set) == 1:
            self.direct.extend(list(on_set))
        else:
            if len(on_diff_set) == 1:
                self.indirect.extend(list(on_diff_set))
        return evidence_views


class OnOnOff(BlicketQuestion):
    def __init__(
        self,
        min_potential_blickets,
        max_potential_blickets,
        min_non_blickets,
        max_non_blickets,
        config_size,
        shuffle,
        rng,
    ):
        super(OnOnOff, self).__init__(
            min_potential_blickets,
            max_potential_blickets,
            min_non_blickets,
            max_non_blickets,
            config_size,
            shuffle,
            rng,
        )

    def get_evidence_views(self):
        # the first non blicket used in habituation and hence the skip
        off_list = self.non_blickets[1:]

        off_view = BlicketView("off")
        off_view.add_objects(off_list)

        first_on, second_on = self.union_sample(self.potential_blickets)

        first_on_view = BlicketView("on")
        first_on_view.add_objects(first_on)

        second_on_view = BlicketView("on")
        second_on_view.add_objects(second_on)

        for on in [first_on, second_on]:
            if len(on) == 1:
                if on[0] not in self.blickets:
                    self.add_blicket(on[0])
        for on in [first_on, second_on]:
            if len(set(on).intersection(self.blickets)) == 0:
                set_blicket = on[:]
                set_blicket.sort()
                set_blicket = tuple(set_blicket)
                if set_blicket not in self.set_blickets:
                    self.add_set_blicket(set_blicket)

        self.add_noise(first_on_view)
        self.add_noise(second_on_view)

        evidence_views = [first_on_view, second_on_view, off_view]

        # bookkeeping for candidate choices
        on_union_set = set(first_on_view.objects).union(second_on_view.objects)
        off_set = set(off_view.objects)

        direct_set = list(off_set.difference(on_union_set))
        screen_off_set = list(off_set.intersection(on_union_set))
        self.direct.extend(direct_set)
        self.screen_off.extend(screen_off_set)

        for on_view in [first_on_view, second_on_view]:
            if len(on_view.objects) == 1:
                obj = on_view.objects[0]
                if obj not in self.direct:
                    self.direct.append(obj)
        for on_view in [first_on_view, second_on_view]:
            diff_set = set(on_view.objects).difference(off_set)
            if len(on_view.objects) > 1 and len(diff_set) == 1:
                obj = diff_set.pop()
                if obj not in self.indirect and obj not in self.direct:
                    self.indirect.append(obj)

        return evidence_views


def serialize(questions):
    question_list = []
    for question in questions:
        view_list = []
        for i in range(6):
            json_dict = {}
            json_dict["light_state"] = question[i].light_state
            json_dict["objects"] = question[i].objects
            view_list.append(json_dict)
        for i in range(6, 8):
            json_dict = {}
            json_dict["light_state"] = question[i][0].light_state
            json_dict["objects"] = question[i][0].objects
            json_dict["label"] = question[i][1]
            json_dict["type"] = question[i][2]
            view_list.append(json_dict)
        for i in range(8, 10):
            json_dict = {}
            json_dict["light_state"] = question[i][0].light_state
            json_dict["objects"] = question[i][0].objects
            json_dict["label"] = question[i][1]
            json_dict["type"] = question[i][2]
            json_dict["ref"] = question[i][3]
            view_list.append(json_dict)
        question_list.append(view_list)
    return question_list


def config_control(size, train, config_size, regime, rng):
    questions = []
    for _ in range(size // 2):
        blicket_machine = OnOffOff(
            ONOFFOFF_MIN_POTENTIAL, ONOFFOFF_MAX_POTENTIAL, ONOFFOFF_MIN_NON, ONOFFOFF_MAX_NON, config_size, True, rng
        )
        context_views = blicket_machine.get_views()

        blicket_machine.sanity_check()
        blicket_machine.check_blickets(context_views)

        cause_questions = blicket_machine.generate_cause_questions(train=train, regime=regime)
        for cause_question in cause_questions:
            blicket_machine.check_labels(cause_question[0], cause_question[1])
        intervention_questions = blicket_machine.generate_intervention_questions(context_views)
        for intervention_question in intervention_questions:
            blicket_machine.check_labels(intervention_question[0], intervention_question[1])
        questions.append(context_views + cause_questions + intervention_questions)
    for _ in range(size // 2):
        blicket_machine = OnOnOff(
            ONONOFF_MIN_POTENTIAL, ONONOFF_MAX_POTENTIAL, ONONOFF_MIN_NON, ONONOFF_MAX_NON, config_size, True, rng
        )
        context_views = blicket_machine.get_views()

        blicket_machine.sanity_check()
        blicket_machine.check_blickets(context_views)

        cause_questions = blicket_machine.generate_cause_questions(train=train, regime=regime)
        for cause_question in cause_questions:
            blicket_machine.check_labels(cause_question[0], cause_question[1])
        intervention_questions = blicket_machine.generate_intervention_questions(context_views)
        for intervention_question in intervention_questions:
            blicket_machine.check_labels(intervention_question[0], intervention_question[1])
        questions.append(context_views + cause_questions + intervention_questions)
    rng.shuffle(questions)
    return questions


def dist_control(size, train, regime, rng):
    questions = []
    for _ in range(size):
        if train:
            blicket_machine = OnOffOff(
                ONOFFOFF_MIN_POTENTIAL,
                ONOFFOFF_MAX_POTENTIAL,
                ONOFFOFF_MIN_NON,
                ONOFFOFF_MAX_NON,
                ALL_CONFIG_SIZE,
                True,
                rng,
            )
        else:
            blicket_machine = OnOnOff(
                ONONOFF_MIN_POTENTIAL,
                ONONOFF_MAX_POTENTIAL,
                ONONOFF_MIN_NON,
                ONONOFF_MAX_NON,
                ALL_CONFIG_SIZE,
                True,
                rng,
            )
        context_views = blicket_machine.get_views()

        blicket_machine.sanity_check()
        blicket_machine.check_blickets(context_views)

        cause_questions = blicket_machine.generate_cause_questions(train=train, regime=regime)
        for cause_question in cause_questions:
            blicket_machine.check_labels(cause_question[0], cause_question[1])
        intervention_questions = blicket_machine.generate_intervention_questions(context_views)
        for intervention_question in intervention_questions:
            blicket_machine.check_labels(intervention_question[0], intervention_question[1])
        questions.append(context_views + cause_questions + intervention_questions)
    rng.shuffle(questions)
    return questions


# Text translation functions
LIGHT_DICT_TEXT = ["off", "undetermined", "on"]
OBJECT_DICT_FILE_PATH = get_data_file_path("acre_objects.json")
with open(OBJECT_DICT_FILE_PATH, "r") as object_dict_file:
    OBJECT_DICT_TEXT = json.load(object_dict_file)


def get_example_text(objects, light_state):
    input_examples = []
    for object in objects:
        obj_desc = OBJECT_DICT_TEXT[str(object)]
        color = obj_desc["color"]
        shape = obj_desc["shape"]
        material = obj_desc["material"]
        # [color, material, shape]
        input_examples.append([color, material, shape])
    return {"input": input_examples, "output": light_state}


# Label has 0, 1, 2 integer which are indexed into the array ["off", "undetermined", "on"]
def get_trial_text(objects, label):
    # object_text = ""
    input_examples = []
    for object in objects:
        obj_desc = OBJECT_DICT_TEXT[str(object)]
        color = obj_desc["color"]
        shape = obj_desc["shape"]
        material = obj_desc["material"]
        # [color, material, shape]
        input_examples.append([color, material, shape])
    return {"input": input_examples, "output": LIGHT_DICT_TEXT[label]}


# Select translation functions
get_example = get_example_text
get_trial = get_trial_text
get_answer_list = lambda: LIGHT_DICT_TEXT


def final_parse(serialized_questions) -> List[Dict[str, Any]]:
    output_data = []
    for sample in serialized_questions:  # For each data in input (6 examples, 4 tests)
        # {"examples": [{"input": [[color, material, shape]], "output": "on"}], "question": {"input": [color, material, shape], "output", "on", "type": "direct"}}
        # examples, ideally is an array of dict(input, output)
        examples = []
        for entry in sample:
            if entry["light_state"] != "no":  # example
                example = get_example(entry["objects"], entry["light_state"])
                examples.append(example)
            else:  # test case
                # question is a dict of input and output
                question = get_trial(entry["objects"], entry["label"])
                output_data.append({"examples": examples, "question": question})

    return output_data



import re
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import sympy
from sympy import Symbol, symbols
from sympy.parsing.sympy_parser import parse_expr

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Make 24 using {numbers}. You can only use each number once. You can use the operators {operators}.
Final answer format instructions:
1. Provide your final answer as a arithmetic expression (no '=' sign).
2. Do not include the target number in the expression.
3. Use '*' for multiplication.
4. Use '/' for division.
"""

DATASET_NAME = "puzzle24"


@dataclass
class Puzzle24Config:
    operators: tuple = ("+", "-", "*", "/")
    min_value: int = 1
    max_value: int = 10
    seed: Optional[int] = None
    size: int = 500

    def validate(self):
        assert len(self.operators) > 0, "At least one operator is required"
        assert self.min_value <= self.max_value, "Minimum value must be less than or equal to maximum value"
        assert self.min_value >= 1, "Minimum value must be at least 1"
        assert self.max_value <= 10, "Maximum value must be at most 10"
        assert self.size > 0, "Size must be greater than 0"


class Puzzle24Dataset(ProceduralDataset):
    def __init__(self, config: Puzzle24Config):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _generate_candidate_expression(self, rng: Random, num_terms: int) -> tuple[sympy.Expr, list[int], list[Symbol]]:
        """Generate a candidate expression with random numbers and operators

        Args:
            rng: Random number generator
            num_terms: Number of terms to include

        Returns:
            Tuple of (sympy expression, list of numbers, list of symbols)
        """
        # Generate random numbers
        numbers = [rng.randint(self.config.min_value, self.config.max_value) for _ in range(num_terms)]

        # Create symbols for building expression
        syms = symbols(f"x:{num_terms}")

        # Build random expression
        expr = syms[0]

        for i in range(1, num_terms):
            op = rng.choice(self.config.operators)
            if op == "+":
                expr = expr + syms[i]
            elif op == "-":
                expr = expr - syms[i]
            elif op == "*":
                expr = expr * syms[i]
            else:
                if numbers[i] != 0:
                    current = int(expr.subs({sym: num for sym, num in zip(syms[:i], numbers[:i])}))
                    remaining = [n for n in numbers[i:] if n != 0]
                    rng.shuffle(remaining)
                    found_divisor = False
                    for div in remaining:
                        if current % div == 0:
                            numbers[i] = div
                            expr = expr / syms[i]
                            found_divisor = True
                            break
                    if not found_divisor:
                        expr = expr - syms[i]
                else:
                    expr = expr + syms[i]

        return expr, numbers, syms

    def __getitem__(self, idx: int) -> dict:
        rng = Random(self.seed + idx)
        while True:
            expr, numbers, syms = self._generate_candidate_expression(rng, 4)
            if expr.subs({sym: num for sym, num in zip(syms, numbers)}) == 24:
                break
        expr_str = str(expr)
        for i, sym in enumerate(syms):
            expr_str = expr_str.replace(str(sym), str(numbers[i]))

        question = QUESTION_TEMPLATE.format(
            numbers=", ".join(map(str, numbers)), operators=", ".join(self.config.operators)
        )
        return {
            "question": question,
            "answer": expr_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "numbers": numbers,
                "difficulty": {"value": (self.config.min_value, self.config.max_value)},
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        reward = 0.01
        if answer is not None:
            try:
                answer = answer.strip()
                user_answer = int(parse_expr(answer))
                solved = user_answer == 24
                used_numbers = [int(num) for num in re.findall(r"\b\d+\b", answer)]
                if len(used_numbers) != 4:
                    reward = 0.01
                elif any(num > self.config.max_value or num < self.config.min_value for num in used_numbers):
                    reward = 0.01
                elif not solved:
                    reward = 0.01
                else:
                    reward = 1.0
            except Exception as e:
                reward = 0.01
        return reward


class Puzzle24Curriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(Puzzle24Curriculum.__name__, Puzzle24Config)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="value",
                levels=[1, 5, 6, 7, 8, 9, 10],
                description="Value of the numbers used in the expression",
                lower_field_name="min_value",
                upper_field_name="max_value",
                ensure_interval=True,
            ),
        )


register_dataset(DATASET_NAME, Puzzle24Dataset, Puzzle24Config, Puzzle24Curriculum)



"""
Code reasing tasks
"""

from .bf import BFConfig, BFCurriculum, BFDataset
from .codeio import CodeIOConfig, CodeIOCurriculum, CodeIODataset

__all__ = [
    "BFConfig",
    "BFDataset",
    "BFCurriculum",
    "CodeIOConfig",
    "CodeIODataset",
    "CodeIOCurriculum",
]



"""Iteratively synthesizes a string by inserting characters according to a pattern.

https://github.com/yongchao98/CodeSteer-v1.0/blob/main/create_dataset/create_dataset_string_synthesis.py
"""

from dataclasses import dataclass
from random import Random
from typing import Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """There are nine different blocks [A] [B] [C] {{A}} {{B}} {{C}} (A) (B) (C)
1. One [A], one [B], and one [C] can be combined to form one {{A}}.
2. One [A] and one [B] can be combined to form one {{C}}.
3. One [B] and one [C] can be combined to form one {{B}}.
4. Two [C] can be combined to form one {{C}}.
5. One {{A}} and one {{C}} can be combined to form one (A) and one (B).
6. Two {{B}} can be combined to form one (C).

Given a certain number of initial blocks, your job is to cycle through the rules 1-6 above, synthesizing new blocks until no more rules can be applied, or until a state (counts of each block type) is repeated.
In the case a state is repeated the answer is the state before the repetition!

The output should be the count of each block type after the rules have been applied in the order they are listed above.
For example 1 0 3 0 2 0 0 0 1 means that you have 1 [A] 0 [B] 3 [C] 0 {{A}} 2 {{B}} 0 {{C}} 0 (A) 0 (B) 1 (C).

Now, you have {A_square} [A], {B_square} [B], and {C_square} [C] blocks. Provide the count of each block type after applying the above rules.
Note: Apply the rules at most {max_iterations} times. If the rules cannot be applied anymore, or if you have reached the maximum number of iterations, stop and provide the current counts.
"""


DATASET_NAME = "string_synthesis"


@dataclass
class StringSynthesisConfig:
    """Configuration for String Synthesis dataset generation"""

    min_initial_blocks: int = 0  # Minimum number of initial blocks
    max_initial_blocks: int = 5  # Maximum number of initial blocks
    max_iterations: int = 1_000  # Maximum number of iterations to apply the rules (Safety check for infinite loops)

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert 0 <= self.min_initial_blocks, "min_initial_blocks must be non-negative"
        assert (
            self.min_initial_blocks <= self.max_initial_blocks
        ), "min_initial_blocks must be less than or equal to max_initial_blocks"
        assert 0 < self.max_iterations, "max_iterations must be positive"


class StringSynthesisDataset(ProceduralDataset):
    """Generates String Synthesis exercises with configurable difficulty"""

    def __init__(self, config: StringSynthesisConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _apply_rule(self, counts: list[int]) -> list[int]:
        """
        Apply the first applicable rule to the given counts.
        In case no rule is applicable, the counts are returned unchanged.
        """
        # label the indices for the counts
        A_square, B_square, C_square, A_curly, B_curly, C_curly, A_round, B_round, C_round = range(9)
        # Rule 1: One [A], one [B], and one [C] can be combined to form one {A}
        if counts[A_square] >= 1 and counts[B_square] >= 1 and counts[C_square] >= 1:
            counts[A_square] -= 1
            counts[B_square] -= 1
            counts[C_square] -= 1
            counts[A_curly] += 1
        # Rule 2: One [A] and one [B] can be combined to form one {C}
        elif counts[A_square] >= 1 and counts[B_square] >= 1:
            counts[A_square] -= 1
            counts[B_square] -= 1
            counts[C_curly] += 1
        # Rule 3: One [B] and one [C] can be combined to form one {B}
        elif counts[B_square] >= 1 and counts[C_square] >= 1:
            counts[B_square] -= 1
            counts[C_square] -= 1
            counts[B_curly] += 1
        # Rule 4: Two [C] can be combined to form one {C}
        elif counts[C_square] >= 2:
            counts[C_square] -= 2
            counts[C_curly] += 1
        # Rule 5: One {A} and one {C} can be combined to form one (A) and one (B)
        elif counts[A_curly] >= 1 and counts[C_curly] >= 1:
            counts[A_curly] -= 1
            counts[C_curly] -= 1
            counts[A_round] += 1
            counts[B_round] += 1
        # Rule 6: Two {B} can be combined to form one (C)
        elif counts[B_curly] >= 2:
            counts[B_curly] -= 2
            counts[C_round] += 1
        return counts

    def _get_answer(self, A_square: int, B_square: int, C_square: int) -> list[list[int]]:
        """Calculate the answer for a given input"""
        # [A] [B] [C] {A} {B} {C} (A) (B) (C)
        counts = [A_square, B_square, C_square] + [0 for _ in range(6)]
        states = [counts]

        for _ in range(self.config.max_iterations):
            new_counts = self._apply_rule(counts[:])
            if new_counts in states:
                break
            states.append(new_counts)
            counts = new_counts

        return states

    def __getitem__(self, idx: int) -> dict:
        """Generate a single String Synthesis question"""
        rng = Random(self.seed + idx)

        A_square = rng.randint(self.config.min_initial_blocks, self.config.max_initial_blocks)
        B_square = rng.randint(self.config.min_initial_blocks, self.config.max_initial_blocks)
        C_square = rng.randint(self.config.min_initial_blocks, self.config.max_initial_blocks)

        states = self._get_answer(A_square, B_square, C_square)
        answer = states[-1]
        answer_str = " ".join(str(x) for x in answer)

        return {
            "question": QUESTION_TEMPLATE.format(
                A_square=A_square,
                B_square=B_square,
                C_square=C_square,
                max_iterations=self.config.max_iterations,
            ),
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "states": states,
                "solution": answer,
                "initial_blocks": (A_square, B_square, C_square),
                "difficulty": {
                    "initial_blocks": (self.config.min_initial_blocks, self.config.max_initial_blocks),
                },
            },
        }


class StringSynthesisCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(StringSynthesisCurriculum.__name__, StringSynthesisConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="initial_blocks",
                levels=[10, 50, 100, 500],
                description="Number of initial blocks",
                lower_field_name="min_initial_blocks",
                upper_field_name="max_initial_blocks",
                ensure_interval=True,
            )
        )


register_dataset(DATASET_NAME, StringSynthesisDataset, StringSynthesisConfig, StringSynthesisCurriculum)



from copy import deepcopy

from .Exceptions import BFSemanticError

functions = dict()  # Global dictionary of function_name --> FunctionCompiler objects


def insert_function_object(function):
    functions[function.name] = function


def get_function_object(name):
    """
    must return a copy of the function
    because we might need to compile function recursively
    and if we don't work on different copies then we will interfere with the current token pointer etc

    for example:
        int increase(int n) { return n+1;}
        int main() {int x = increase(increase(1));}

    while compiling the first call, we start a compilation of the same function object in the second call
    """
    return deepcopy(functions[name])


def check_function_exists(function_token, parameters_amount):
    function_name = function_token.data
    if function_name not in functions:
        raise BFSemanticError("Function '%s' is undefined" % str(function_token))

    function = functions[function_name]
    if len(function.parameters) != parameters_amount:
        raise BFSemanticError(
            "Function '%s' has %s parameters (called it with %s parameters)"
            % (str(function_token), len(function.parameters), parameters_amount)
        )



from .Functions import insert_function_object
from .Token import Token


class LibraryFunctionCompiler:
    def __init__(self, name, type, parameters, code):
        self.name = name
        self.type = type
        self.parameters = parameters
        self.code = code

    def get_code(self, current_stack_pointer):
        return self.code


def get_readint_code():
    # res, tmp, input, loop
    # tmp is used for multiplication
    """
    res = 0
    loop = 1

    while loop
        loop = 0
        input = input()
        if input != newline # todo add a eof check as well. run it in several interpreters to look for common ways for "end of number" input
            loop = 1
            res *= 10 + char_to_digit(input)
    """

    code = "[-]"  # clear res = 0
    code += ">[-]"  # tmp = 0
    code += ">>[-]+"  # loop = 1

    code += "["  # while loop == 1
    code += "[-]"  # loop = 0
    code += "<"  # point to input
    code += ","  # input character
    code += "----------"  # sub 10 (check for newline)

    code += "["  # if input is not newline
    code += ">"  # point to loop
    code += "+"  # loop = 1

    # multiply res by 10 and add the input digit
    code += "<<<"  # point to res
    code += "[>+<-]"  # move res to tmp
    code += ">"  # point to tmp
    code += "[<++++++++++>-]"  # res = tmp * 10, tmp = 0
    code += ">"  # point to input
    code += "-" * (
        0x30 - 10
    )  # convert character to a digit by subtracting 0x30 from it (we already subtracted 10 before)
    code += "[<<+>>-]"  # res += input
    code += "]"  # end if

    code += ">"  # point to loop
    code += "]"  # end while

    code += "<<<"  # point to res

    return code


def get_printint_code():
    # return_cell value_to_print_cell

    code = ">"  # point to value_to_print cell
    code += ">[-]" * 8 + "<" * 8  # zero some cells

    code += ">++++++++++<"  # div amount
    code += "[->-[>+>>]>[+[<+>-]>+>>]<<<<<]"  # value_to_print/10
    code += ">[-]"  # zero d-n%d
    code += ">>"  # point to div result

    code += ">++++++++++<"  # div amount
    code += "[->-[>+>>]>[+[<+>-]>+>>]<<<<<]"  # res/10
    code += ">[-]"  # zero d-n%d
    code += ">>"  # point to div result

    code += "["  # if the first digit is not 0
    code += ">++++++[<++++++++>-]<."  # add 48 to the first digit and print it
    code += "<<"
    code += "+>"  # set is_over_100 to true
    code += "+>"  # add 1 to the second digit so it prints even when it's 0
    code += "[-]"  # zero the first digit
    code += "]"  # end if

    code += "<"  # point to the second digit

    code += "["  # if the second digit is not 0
    code += "<[>-<-]"  # if is_over_100 is true then subtract 1 from the second digit
    code += "++++++[>++++++++<-]>."  # add 48 to the second digit and print it
    code += "[-]"  # zero the second digit
    code += "]"  # end if

    code += "<<"  # point to the cell after the third digit
    code += "++++++[<++++++++>-]<."  # add 48 to the third digit and print it
    code += "[-]"  # zero the third digit
    code += "<<"  # point to value_to_print_cell which is 0

    code += "<"  # point to return_cell
    return code


def get_readchar_code():
    # read input into "return value cell". no need to move the pointer
    code = ","
    return code


def get_printchar_code():
    # point to parameter, output it, and then point back to "return value cell"
    code = ">.<"
    return code


def insert_library_functions():
    readint = LibraryFunctionCompiler("readint", Token.INT, list(), get_readint_code())
    insert_function_object(readint)

    printint = LibraryFunctionCompiler("printint", Token.VOID, [Token.INT], get_printint_code())
    insert_function_object(printint)

    readchar = LibraryFunctionCompiler("readchar", Token.INT, list(), get_readchar_code())
    insert_function_object(readchar)

    printchar = LibraryFunctionCompiler("printchar", Token.VOID, [Token.INT], get_printchar_code())
    insert_function_object(printchar)



# -*- coding: utf-8 -*-


ONOFFOFF_MIN_POTENTIAL = 2
ONOFFOFF_MAX_POTENTIAL = 3
ONOFFOFF_MIN_NON = 3
ONOFFOFF_MAX_NON = 5

ONONOFF_MIN_POTENTIAL = 2
ONONOFF_MAX_POTENTIAL = 4
ONONOFF_MIN_NON = 3
ONONOFF_MAX_NON = 4

ATTR_CONFIG_SIZE = 24
ALL_CONFIG_SIZE = 48



import re
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from magiccube.cube import Cube, CubeMove, CubeMoveType
from magiccube.solver.basic.basic_solver import BasicSolver

from ..coaching import BaseCurriculum, RangeAttributeDefinition, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "rubiks_cube"


@dataclass
class RubiksCubeConfig:
    """Configuration for RubiksCube task generation"""

    min_scramble_steps: int = 3  # Minimum number of random steps from initial state
    max_scramble_steps: int = 10  # Maximum number of random steps from initial state
    cube_size: int = 3  # Default to a standard 3x3x3 cube
    remove_ansi: bool = True
    seed: Optional[int] = None
    size: int = 500

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.cube_size > 1, "cube_size must be greater than 1"
        assert self.cube_size < 7, "cube_size must be less than 7"
        assert self.min_scramble_steps > 0, "min_scramble_steps must be greater than 0"
        assert (
            self.max_scramble_steps >= self.min_scramble_steps
        ), "max_scramble_steps must be greater than min_scramble_steps"


class RubiksCubeDataset(ProceduralDataset):
    """Generates RubiksCube tasks"""

    def __init__(self, config: RubiksCubeConfig):
        self._prompt_templates = [
            "You are given a {cube_size}x{cube_size}x{cube_size} Rubik's cube. It looks like this:\n\n{cube_render} \n\nPlease provide a solution to solve this cube using Singmaster notation. Do not combine any steps, for instance, do not write 'U2', and instead write 'U U'.",
            "You see a size {cube_size} Rubik's cube arranged as follows::\n\n{cube_render} \n\nPlease provide a solution to solve this cube using Singmaster notation. Do not combine any steps, for instance, do not write 'U2', and instead write 'U U'.",
        ]
        super().__init__(config=config, seed=config.seed, size=config.size)

    def _generate_random_moves(self, rng: Random, cube: Cube, num_steps: int = 50, wide=None) -> list[CubeMove]:
        """Generate a list of random moves (but don't apply them).
        By default scramble only uses wide moves to cubes with size >=4."""

        if wide is None and cube.size <= 3:
            wide = False
        elif wide is None and cube.size > 3:
            wide = True

        possible_moves = [
            CubeMoveType.L,
            CubeMoveType.R,  # CubeMoveType.M,
            CubeMoveType.D,
            CubeMoveType.U,  # CubeMoveType.E,
            CubeMoveType.B,
            CubeMoveType.F,  # CubeMoveType.S,
        ]
        movements = [
            CubeMove(
                rng.choice(possible_moves),
                rng.choice([False, True]),  # reversed
                rng.choice([False, True]) if wide else False,  # wide
                rng.randint(1, cube.size // 2) if wide else 1,  # layer
            )
            for _ in range(num_steps)
        ]

        return movements

    def __getitem__(self, idx: int) -> dict:
        """Generate a single RubiksCube task

        Returns:
            dict with keys:
                - question: str, the task description with cube string
                - answer: None, indicating to use the dynamic evaluator
                - metadata: dict with generation parameters and example solution
        """
        rng = Random(self.seed + idx)

        cube = Cube(self.config.cube_size)
        num_steps = rng.randint(self.config.min_scramble_steps, self.config.max_scramble_steps)
        scramble_moves = self._generate_random_moves(rng, cube, num_steps=num_steps)
        cube.rotate(scramble_moves)

        # render cube
        if self.config.remove_ansi:
            cube_render = self.remove_ansi(str(cube))
        else:
            cube_render = str(cube)

        if self.config.cube_size == 3:
            solver = BasicSolver(cube)
            actions = solver.solve()
            actions_string = " ".join([str(move) for move in actions])
        else:
            actions = None
            actions_string = ""

        return {
            "question": rng.choice(self._prompt_templates).format(
                cube_size=self.config.cube_size, cube_render=cube_render
            ),
            "answer": None,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "cube_size": self.config.cube_size,
                "scramble_steps": num_steps,
                "scramble_moves": " ".join([str(move) for move in scramble_moves]),
                "example_correct_answer": actions_string,
                "difficulty": {
                    "cube_size": self.config.cube_size,
                    "scramble_steps": (self.config.min_scramble_steps, self.config.max_scramble_steps),
                },
            },
        }

    def partial_score(self, cube: Cube) -> float:
        """
        Returns a fraction between 0 and 1, indicating how many stickers are
        correctly positioned (i.e., match the solved color for that face).
        """
        total_stickers = 6 * (cube.size**2)
        correct_stickers = 0

        for face_index in range(6):
            face = cube.faces[face_index]

            solved_color = face[cube.size // 2][cube.size // 2].color
            for row in range(cube.size):
                for col in range(cube.size):
                    sticker = face[row][col]
                    if sticker.color == solved_color:
                        correct_stickers += 1

        return correct_stickers / total_stickers

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided solves the cube, with partial rewards."""
        reward = 0.0  # default
        if answer is not None:
            eval_cube = Cube(entry["metadata"]["cube_size"])
            eval_cube.rotate(entry["metadata"]["scramble_moves"])
            try:
                expanded_answer = self.expand_moves(answer)
                eval_cube.rotate(expanded_answer)

                # 3) Check if fully solved
                solved = eval_cube.is_done()
                if solved:
                    reward = 1.0
                else:
                    partial = self.partial_score(eval_cube)

                    if len(answer.strip()) > 0:
                        reward = max(0.05, partial)
                    else:
                        reward = max(0.01, partial)
            except:
                reward = 0.01
        return reward

    def remove_ansi(self, line):
        """Remove terminal colors from magiccube rendering"""
        ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
        return ansi_escape.sub("", line)

    def expand_moves(self, move_str):
        try:
            moves = move_str.split()
            expanded = []
            for move in moves:
                # Split the move into the base part and any trailing digits
                match = re.fullmatch(r"^([^\d]*)(\d*)$", move)
                if match:
                    base, num_part = match.groups()
                    if num_part:
                        expanded.extend([base] * int(num_part))
                    else:
                        expanded.append(base)
            return " ".join(expanded).strip()
        except Exception as e:
            return move_str


class RubiksCubeCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(RubiksCubeCurriculum.__name__, RubiksCubeConfig)

        # Define attributes
        self._define_attributes(
            ScalarAttributeDefinition(
                name="cube_size",
                field_name="cube_size",
                levels=[3, 4, 5, 6, 7],
                description="Board size",
            ),
            RangeAttributeDefinition(
                name="scramble_steps",
                levels=[3, 10, 25, 50, 100],
                description="Number of random moves to scramble the cube",
                lower_field_name="min_scramble_steps",
                upper_field_name="max_scramble_steps",
                ensure_interval=True,
            ),
        )


# Register the dataset
register_dataset(DATASET_NAME, RubiksCubeDataset, RubiksCubeConfig, RubiksCubeCurriculum)



"""Number sorting task generator"""

import json
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import numpy as np

from ..coaching import BaseCurriculum, RangeAttributeDefinition, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "number_sorting"


@dataclass
class NumberSortingConfig:
    """Configuration for number sorting task generation"""

    min_numbers: int = 3  # Minimum numbers to sort
    max_numbers: int = 10  # Maximum numbers to sort
    min_decimals: int = 0  # Minimum decimal places
    max_decimals: int = 2  # Maximum decimal places
    min_value: float = -100.0  # Minimum value
    max_value: float = 100.0  # Maximum value
    seed: Optional[int] = None
    size: int = 500  # Virtual dataset size

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.min_numbers > 0, "min_numbers must be positive"
        assert self.min_numbers <= self.max_numbers, "max_numbers must be >= min_numbers"
        assert self.min_decimals >= 0, "min_decimals must be non-negative"
        assert self.min_decimals <= self.max_decimals, "max_decimals must be >= min_decimals"
        assert self.min_value < self.max_value, "max_value must be > min_value"


class NumberSortingDataset(ProceduralDataset):
    """Generates number sorting tasks"""

    def __init__(self, config: NumberSortingConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)
        self.added_instruction = """
Please follow the instruction below:
## 1. Let all your answers be a list of numbers. Instead of reporting your answer as -69, -13, 1, 7, 11, 43, 59, 61, use ['-69', '-13', '1', '7', '11', '43', '59', '61'] instead
## 2. Convert all numbers in the square brackets as strings. For example, ['-69', '-13', '1', '7', '11', '43', '59', '61']
"""

    def _generate_numbers(self, rng: Random, count: int) -> tuple[list[float], list[str]]:
        """Generate list of numbers and their string representations"""
        numbers = []
        number_strs = []

        for _ in range(count):
            num = rng.uniform(self.config.min_value, self.config.max_value)
            decimals = rng.randint(self.config.min_decimals, self.config.max_decimals)
            num = np.round(num, decimals)
            numbers.append(num)
            number_strs.append(str(num))

        return numbers, number_strs

    def __getitem__(self, idx: int) -> dict:
        """Generate a single sorting task"""
        rng = Random(self.seed + idx)

        count = rng.randint(self.config.min_numbers, self.config.max_numbers)
        numbers, number_strs = self._generate_numbers(rng, count)

        # Generate both ascending and descending answers
        asc_numbers = sorted(numbers)
        desc_numbers = sorted(numbers, reverse=True)

        # Format answers as string lists
        asc_answer = [str(n) for n in asc_numbers]
        desc_answer = [str(n) for n in desc_numbers]

        # Randomly choose ascending or descending
        is_ascending = rng.choice([True, False])
        direction = "ascending" if is_ascending else "descending"
        answer = asc_answer if is_ascending else desc_answer
        question = f"Sort these numbers in {direction} order: {', '.join(number_strs)}" + self.added_instruction

        return {
            "question": question,
            "answer": str(answer),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "original_numbers": number_strs,
                "direction": direction,
                "sorted_numbers": answer,
                "numbers": count,
                "difficulty": {
                    "numbers": (self.config.min_numbers, self.config.max_numbers),
                    "decimals": (self.config.min_decimals, self.config.max_decimals),
                    "value": (self.config.min_value, self.config.max_value),
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Score the user's answer against the expected answer.

        Args:
            answer (Optional[str]): The user's answer string.
            entry (dict[str, Any]): The original dataset entry with the correct answer.

        Returns:
            float: 1.0 for a correct answer, 0.0 for incorrect.
        """
        if answer is None:
            return 0.0

        try:
            # Try to parse the user's answer as a JSON list first
            try:
                answer = answer.replace("'", '"')
                user_answer = json.loads(answer)
            except json.JSONDecodeError:
                return 0.0  # JSON parsing failed

            if not isinstance(user_answer, list):
                return 0.0

            # Get the expected answer
            try:
                expected_answer = json.loads(entry["answer"])
            except json.JSONDecodeError:
                # Fall back to eval if necessary
                expected_answer = eval(entry["answer"])

            # Check if the lists have the same length
            if len(user_answer) != len(expected_answer):
                return 0.0

            # Convert both answers to floats for comparison
            user_floats = [float(num) for num in user_answer]
            expected_floats = [float(num) for num in expected_answer]

            # First, verify the user's answer is properly sorted
            direction = entry["metadata"]["direction"]
            is_correctly_sorted = False

            if direction == "ascending":
                is_correctly_sorted = user_floats == sorted(user_floats)
            else:  # descending
                is_correctly_sorted = user_floats == sorted(user_floats, reverse=True)

            if not is_correctly_sorted:
                return 0.0

            # Check if the values are close enough (allowing for small rounding differences)
            tolerance = 1  # Increased tolerance to handle decimal differences
            for i in range(len(user_floats)):
                if abs(user_floats[i] - expected_floats[i]) > tolerance:
                    return 0.0

            return 1.0
        except Exception:
            # Any parsing error means the answer is incorrect
            return 0.0


class NumberSortingCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(NumberSortingCurriculum.__name__, NumberSortingConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="numbers",
                levels=[10, 100, 500, 1000],
                description="How many numbers to sort",
                lower_field_name="min_numbers",
                upper_field_name="max_numbers",
                ensure_interval=True,
            ),
            RangeAttributeDefinition(
                name="decimals",
                levels=list(range(0, 8)),
                description="Number of decimal places",
                lower_field_name="min_decimals",
                upper_field_name="max_decimals",
                ensure_interval=True,
            ),
            ScalarAttributeDefinition(
                name="min_value",
                field_name="min_value",
                levels=[-100, -500, -1000, -10000],
                description="Minimum number value",
            ),
            ScalarAttributeDefinition(
                name="max_value",
                field_name="max_value",
                levels=[100, 500, 1000, 10000],
                description="Maximum number value",
            ),
        )


register_dataset(DATASET_NAME, NumberSortingDataset, NumberSortingConfig, NumberSortingCurriculum)



from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "self_reference"


def is_prime(n):
    """Return True if n is a prime number, False otherwise."""
    if n < 2:
        return False
    for i in range(2, int(n**0.5) + 1):
        if n % i == 0:
            return False
    return True


def is_composite(n):
    """
    Return True if n is composite.
    (Composite means an integer greater than 1 that is not prime.)
    """
    return n > 1 and not is_prime(n)


def generate_dynamic_puzzle(difficulty, rng):
    """
    Dynamically generates a 7-statement self-referential puzzle.

    The seven statements (with parameters determined by this function) are:

      1. "At least a of these 7 statements are true."
      2. "At most b of these 7 statements are false."
      3. "Exactly c of these 7 statements are true."
      4. "Exactly d of these 7 statements are false."
      5. "Either Statement 3 or Statement 4 is true, but not both."
      6. "The number of true statements is a prime number."
      7. "The number of false statements is a composite number."

    The idea is to choose an intended number T (1 ≤ T ≤ 6) of true statements
    and then “plant” an intended solution. In our construction the truth values
    for Statements 6 and 7 are forced by T (e.g. Statement 6 should be true exactly
    when T is prime). For the first four statements the numeric parameters (a, b, c, d)
    are chosen so that the statement evaluates correctly when compared to T.

    The difficulty parameter (an integer, e.g. 1 for easy up to 10 for hard)
    influences how “borderline” the numeric choices are. At lower difficulty the numbers
    are chosen with a clear gap; at higher difficulty they are chosen closer to T.

    Returns:
        dict: A puzzle dictionary containing:
              - 'n': number of statements (always 7 here),
              - 'statements_text': a list of 7 strings (one per statement),
              - 'parameters': a dict with the numeric parameters (for statements 1-4),
              - 'intended_assignment': the intended truth values (list of 7 booleans),
              - 'intended_T': the intended number of true statements.
    """
    n = 7

    # Choose an intended number of true statements, T, from 1 to 6 (nontrivial).
    T = rng.choice(range(1, n))

    # For the global statements (6 and 7), the intended truth is forced:
    intended6 = is_prime(T)  # Statement 6 must be true if T is prime.
    intended7 = is_composite(n - T)  # Statement 7 must be true if (# false) is composite.

    # Among statements 1-5, we need exactly k trues such that overall the total becomes T.
    # Let k = T - (truth from statements 6 and 7).
    forced_true_count = (1 if intended6 else 0) + (1 if intended7 else 0)
    k = T - forced_true_count
    # k must be between 0 and 5.
    if not (0 <= k <= 5):
        # If for some reason it is not in range, fall back to a known configuration (T=4).
        T = 4
        intended6 = False
        intended7 = False
        k = 4  # so that overall T=4.
        intended_assignment_15 = [True, True, True, True, False]
    else:
        # For statements 1-5, randomly choose which ones are intended true.
        # We'll index these as 0..4 corresponding to statements 1..5.
        intended_assignment_15 = [False] * 5
        if k > 0:
            true_indices = set(rng.sample(range(5), k))
            for i in true_indices:
                intended_assignment_15[i] = True

    # Now, for statements 1-4, choose numeric parameters based on whether the statement is
    # intended to be true or false. We use the difficulty parameter to control the "margin."
    #
    # For statement 1: "At least a of these 7 statements are true."
    # The condition is: T >= a.
    def choose_at_least_param(T, intended, diff, rng):
        # diff will be used as a margin factor: lower diff => wider gap.
        if intended:  # must have a <= T.
            # At easy difficulty, choose a clearly below T (if possible).
            low = 1
            high = T
            # At lower difficulty, bias toward the lower end.
            return rng.randint(low, high)
        else:  # must have a > T.
            low = T + 1
            high = n  # a can be at most n.
            if low > high:
                return n
            return rng.randint(low, high)

    a_param = choose_at_least_param(T, intended_assignment_15[0], difficulty, rng)

    # For statement 2: "At most b of these 7 statements are false."
    # F = n - T, so condition is: (n - T) <= b   <=>   T >= n - b.
    def choose_at_most_param(T, intended, diff, rng):
        if intended:  # b must be >= n - T.
            low = n - T
            high = n
            return rng.randint(low, high)
        else:
            # b must be < n - T.
            low = 0
            high = max(n - T - 1, 0)
            return rng.randint(low, high)

    b_param = choose_at_most_param(T, intended_assignment_15[1], difficulty, rng)

    # For statement 3: "Exactly c of these 7 statements are true."
    def choose_exactly_true_param(T, intended, diff, rng):
        if intended:
            return T
        else:
            choices = [x for x in range(0, n + 1) if x != T]
            return rng.choice(choices)

    c_param = choose_exactly_true_param(T, intended_assignment_15[2], difficulty, rng)

    # For statement 4: "Exactly d of these 7 statements are false."
    # Condition: (n - T) == d.
    def choose_exactly_false_param(T, intended, diff, rng):
        false_count = n - T
        if intended:
            return false_count
        else:
            choices = [x for x in range(0, n + 1) if x != false_count]
            return rng.choice(choices)

    d_param = choose_exactly_false_param(T, intended_assignment_15[3], difficulty, rng)

    # For statement 5: "Either Statement 3 or Statement 4 is true, but not both."
    # We do not need a parameter here; the intended condition is that the truth values for
    # statements 3 and 4 (which are positions 2 and 3 in our 0-indexed list) differ.
    # The intended truth for statement 5 is taken from our assignment.
    # (Later the verification function will check: solution[2] != solution[3].)

    # Build the intended assignment for all 7 statements.
    # For statements 1-5, we use our generated intended_assignment_15.
    intended_assignment = [
        intended_assignment_15[0],
        intended_assignment_15[1],
        intended_assignment_15[2],
        intended_assignment_15[3],
        intended_assignment_15[4],
        intended6,
        intended7,
    ]

    # (If the total intended true count doesn't equal T, adjust statement 5.)
    current_T = sum(intended_assignment)
    if current_T != T:
        # Since only statement 5 is free (its parameter wasn't numeric),
        # force its intended truth to be what is needed.
        intended_assignment[4] = T - (current_T - (1 if intended_assignment[4] else 0)) == 1

    # Now build the text for each statement.
    statements_text = [
        f"Statement 1: 'At least {a_param} of these 7 statements are true.'",
        f"Statement 2: 'At most {b_param} of these 7 statements are false.'",
        f"Statement 3: 'Exactly {c_param} of these 7 statements are true.'",
        f"Statement 4: 'Exactly {d_param} of these 7 statements are false.'",
        "Statement 5: 'Either Statement 3 or Statement 4 is true, but not both.'",
        "Statement 6: 'The number of true statements is a prime number.'",
        "Statement 7: 'The number of false statements is a composite number.'",
    ]

    return {
        "n": n,
        "statements_text": statements_text,
        "parameters": {
            "a": a_param,
            "b": b_param,
            "c": c_param,
            "d": d_param,
        },
        "intended_assignment": intended_assignment,
        "intended_T": T,
        "difficulty": difficulty,
    }


def verify_solution_dynamic(puzzle, solution):
    """
    Verifies a candidate solution for a dynamically generated puzzle.

    The rules are:
      - If a statement is marked True, then its claim must hold.
      - If a statement is marked False, then its claim must fail.

    The conditions are as follows:
      1. "At least a of these 7 statements are true."  => (T >= a)
      2. "At most b of these 7 statements are false."   => (F <= b)
      3. "Exactly c of these 7 statements are true."    => (T == c)
      4. "Exactly d of these 7 statements are false."   => (F == d)
      5. "Either Statement 3 or Statement 4 is true, but not both." => (solution[2] != solution[3])
      6. "The number of true statements is a prime number." => is_prime(T)
      7. "The number of false statements is a composite number." => is_composite(F)

    Parameters:
       puzzle (dict): The puzzle dictionary returned by generate_dynamic_puzzle.
       solution (list of bool): A candidate assignment (length 7).

    Returns:
       bool: True if candidate is self-consistent; False otherwise.
    """
    n = puzzle["n"]
    if len(solution) != n:
        return False
    T = sum(solution)
    F = n - T
    params = puzzle["parameters"]

    # Statement 1: "At least a of these 7 statements are true."
    cond1 = T >= params["a"]
    if solution[0] and not cond1:
        return False
    if not solution[0] and cond1:
        return False

    # Statement 2: "At most b of these 7 statements are false."
    cond2 = F <= params["b"]
    if solution[1] and not cond2:
        return False
    if not solution[1] and cond2:
        return False

    # Statement 3: "Exactly c of these 7 statements are true."
    cond3 = T == params["c"]
    if solution[2] and not cond3:
        return False
    if not solution[2] and cond3:
        return False

    # Statement 4: "Exactly d of these 7 statements are false."
    cond4 = F == params["d"]
    if solution[3] and not cond4:
        return False
    if not solution[3] and cond4:
        return False

    # Statement 5: "Either Statement 3 or Statement 4 is true, but not both."
    cond5 = solution[2] != solution[3]
    if solution[4] and not cond5:
        return False
    if not solution[4] and cond5:
        return False

    # Statement 6: "The number of true statements is a prime number."
    cond6 = is_prime(T)
    if solution[5] and not cond6:
        return False
    if not solution[5] and cond6:
        return False

    # Statement 7: "The number of false statements is a composite number."
    cond7 = is_composite(F)
    if solution[6] and not cond7:
        return False
    if not solution[6] and cond7:
        return False

    return True


def print_puzzle_dynamic(puzzle):
    """Prints the dynamically generated puzzle."""
    x = ""
    for stmt in puzzle["statements_text"]:
        x = x + " - " + stmt + "\n"
    return x


def solve_puzzle_dynamic(puzzle):
    """
    Searches all 2^7 possible truth assignments and returns those that
    are self-consistent with the generated puzzle.
    """
    n = puzzle["n"]
    valid_solutions = []
    for i in range(2**n):
        candidate = [(i >> j) & 1 == 1 for j in range(n)]
        if verify_solution_dynamic(puzzle, candidate):
            valid_solutions.append(candidate)
    return valid_solutions


@dataclass
class SelfReferenceConfig:
    """Configuration for SelfReference puzzle generation"""

    difficulty: int = 5
    seed: Optional[int] = None
    size: int = 500

    def validate(self):
        """Validate configuration parameters"""
        assert 1 <= self.difficulty <= 10, "difficulty must be between 1 and 10"


class SelfReferenceDataset(ProceduralDataset):
    """Generates self-referential puzzles"""

    def __init__(self, config: SelfReferenceConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single SelfReference task

        Returns:
            dict with keys:
                - question: str, the task description
                - answer: str, a solution string
                - metadata: dict with generation parameters
        """
        rng = Random(self.seed + idx)
        difficulty = self.config.difficulty

        # Generate puzzle
        puzzle = generate_dynamic_puzzle(difficulty, rng)
        puzz_s = (
            "Given the truthfulness of these statements, please tell me the number of possible solutions: \n"
            + print_puzzle_dynamic(puzzle)
        )

        # Solve puzzle
        solutions = solve_puzzle_dynamic(puzzle)
        answer = str(len(solutions))

        return {
            "question": puzz_s,
            "answer": answer,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "difficulty": {"difficulty": difficulty},
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided solves the SelfReference task.

        The function awards 1.0 for a correct answer.

        Args:
            answer (Optional[str]): The user's answer.
            entry (dict[str, Any]): The original dataset entry containing the correct answer.

        Returns:
            float: The computed score between 0.0 and 1.0.
        """

        if isinstance(answer, str):
            if answer == str(entry["answer"]):
                return 1.0  # Yay
        return 0.0


class SelfReferenceCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(SelfReferenceCurriculum.__name__, SelfReferenceConfig)
        self._define_attributes(
            ScalarAttributeDefinition(
                name="difficulty",
                field_name="difficulty",
                levels=list(range(1, 11)),
                description="The difficulty of the puzzle",
            )
        )


register_dataset(DATASET_NAME, SelfReferenceDataset, SelfReferenceConfig, SelfReferenceCurriculum)



"""N Queens puzzle generator

A generalization of the 8-queens puzzle to any board size.
https://en.wikipedia.org/wiki/Eight_queens_puzzle
"""

from copy import deepcopy
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

MIN_BOARD_SIZE = 4
MAX_BOARD_SIZE = 12

QUESTION_TEMPLATE = """Your job is to complete an n x n chess board with n Queens in total, such that no two attack each other.

No two queens attack each other if they are not in the same row, column, or diagonal.

You can place a queen by replacing an underscore (_) with a Q.

Your output should be also a board in the same format as the input, with queens placed on the board by replacing underscores with the letter Q.

Given the below board of size {n} x {n} your job is to place {num_removed} queen(s) on the board such that no two queens attack each other.
{puzzle}
"""

DATASET_NAME = "n_queens"


@dataclass
class NQueensConfig:
    """Configuration for N Queens puzzle generation"""

    n: int = 8  # Board size
    min_remove: int = 1  # Minimum number of queens to remove from solved board
    max_remove: int = 7  # Maximum number of queens to remove from solved board

    size: int = 500  # Virtual dataset size
    seed: Optional[int] = None

    def validate(self):
        """Validate configuration parameters"""
        assert MIN_BOARD_SIZE <= self.n <= MAX_BOARD_SIZE, f"n must be between {MIN_BOARD_SIZE} and {MAX_BOARD_SIZE}"
        assert 1 <= self.min_remove <= self.max_remove, "min_remove must be between 1 and max_remove"
        assert self.min_remove <= self.max_remove <= self.n, "max_remove must be between min_remove and n"


class NQueensDataset(ProceduralDataset):
    """Generates N Queens puzzles with configurable difficulty"""

    def __init__(self, config: NQueensConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)
        self._solutions = self._get_all_solutions(config.n)

    def _get_all_solutions(self, n: int) -> list[list[list[str]]]:
        """Get all solutions for the N Queens puzzle"""

        visited_cols = set()
        visited_pos_diag = set()
        visited_neg_diag = set()

        res = []
        board = [["_"] * n for _ in range(n)]

        def backtrack(row: int):
            if row == n:
                res.append(deepcopy(board))
                return

            for col in range(n):
                if col in visited_cols or (row + col) in visited_pos_diag or (row - col) in visited_neg_diag:
                    continue

                visited_cols.add(col)
                visited_pos_diag.add(row + col)
                visited_neg_diag.add(row - col)
                board[row][col] = "Q"
                backtrack(row + 1)
                visited_cols.remove(col)
                visited_pos_diag.remove(row + col)
                visited_neg_diag.remove(row - col)
                board[row][col] = "_"

        backtrack(0)
        return res

    def _create_puzzle(self, solved_board: list[list[str]], num_removed: int, rng: Random) -> list[list[str]]:
        """Create puzzle by removing queens from solved board"""
        puzzle = deepcopy(solved_board)
        queens = [(i, j) for i in range(len(puzzle)) for j in range(len(puzzle)) if puzzle[i][j] == "Q"]
        rng.shuffle(queens)
        for i in range(num_removed):
            x, y = queens[i]
            puzzle[x][y] = "_"
        return puzzle

    def _board_to_string(self, board: list[list[str]]) -> str:
        """Convert board to string representation"""
        return "\n".join(" ".join(x for x in row) for row in board)

    def _string_to_board(self, board_str: str) -> list[list[str]]:
        """Convert string representation to board"""
        return [list(row.split()) for row in board_str.strip().split("\n")]

    def _is_tractable_solution(self, puzzle: list[list[str]], solution: list[list[str]]) -> bool:
        """Check if a solution is achievable from the starting state of the puzzle"""
        for r in range(len(puzzle)):
            for c in range(len(puzzle)):
                if puzzle[r][c] == "Q" and solution[r][c] != "Q":
                    return False
        return True

    def __getitem__(self, idx: int) -> dict:
        """Generate a single N Queens puzzle"""
        rng = Random(self.seed + idx)

        # Randomly select a valid solution
        solved_board = rng.choice(self._solutions)

        # Create puzzle by removing queens
        num_removed = rng.randint(self.config.min_remove, self.config.max_remove)
        puzzle = self._create_puzzle(solved_board, num_removed, rng)
        puzzle_str = self._board_to_string(puzzle)

        # Filter all solutions that are intractable from the puzzle's starting state
        valid_solutions = [board for board in self._solutions if self._is_tractable_solution(puzzle, board)]
        valid_solutions_str = sorted({self._board_to_string(board) for board in valid_solutions})

        return {
            "question": QUESTION_TEMPLATE.format(puzzle=puzzle_str, n=len(puzzle), num_removed=num_removed),
            "answer": rng.choice(valid_solutions_str),  # choose arbitary answer (e.g. for SFT)
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "puzzle": puzzle,
                "solutions": valid_solutions,
                "num_removed": num_removed,
                "valid_answers": valid_solutions_str,
                "difficulty": {
                    "n": self.config.n,
                    "num_removed": (self.config.min_remove, self.config.max_remove),
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        if isinstance(answer, str):
            valid_solutions = entry["metadata"]["valid_answers"]
            if answer in valid_solutions:
                return 1.0
            try:
                answer = self._board_to_string(eval(answer))
                if answer in valid_solutions:
                    return 0.5
            except Exception as e:
                pass
        return 0.0


class NQueensCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(NQueensCurriculum.__name__, NQueensConfig)

        self._define_attributes(
            ScalarAttributeDefinition(
                name="n",
                field_name="n",
                levels=[4, 6, 8, 10, 12],
                description="Board size",
            ),
            RangeAttributeDefinition(
                name="num_removed",
                levels=[2, 4, 6, 8, 10],
                description="Number of queens to remove",
                lower_field_name="min_remove",
                upper_field_name="max_remove",
                ensure_interval=True,
            ),
        )


register_dataset(DATASET_NAME, NQueensDataset, NQueensConfig, NQueensCurriculum)



from .arc_1d import Arc1DConfig, Arc1DCurriculum, Arc1DDataset
from .arc_agi import ArcAgiConfig, ArcAgiCurriculum, ArcAgiDataset
from .rearc import ReArcConfig, ReArcCurriculum, ReArcDataset

__all__ = [
    "Arc1DConfig",
    "Arc1DDataset",
    "Arc1DCurriculum",
    "ArcAgiConfig",
    "ArcAgiDataset",
    "ArcAgiCurriculum",
    "ReArcDataset",
    "ReArcConfig",
    "ReArcCurriculum",
]



import json
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import pyfiglet

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..data import get_data_file_path
from ..factory import ProceduralDataset, register_dataset

# These ones are funky and probably aren't good for train/testing
BAD_FONTS = [
    "pyramid",
    "runyc",
    "assalt_m",
    "term",
    "tengwar",
    "heart_right",
    "faces_of",
    "heroboti",
    "hieroglyphs",
    "rainbow_",
    "notie_ca",
    "ghost",
    "rampage_",
    "atc_____",
    "pacos_pe",
    "mad_nurs",
    "icl-1900",
    "joust___",
    "dcs_bfmo",
    "letter_w",
    "flyn_sh",
    "fun_face",
    "morse2",
    "tecrvs__",
    "ntgreek",
    "tsalagi",
    "etcrvs__",
    "faces_of",
    "future_8",
    "efti_robot",
    "danc4",
    "p_s_h_m_",
    "smkeyboard",
    "konto",
    "odel_lak",
    "courb",
    "jerusalem",
    "nfi1____",
    "keyboard",
    "konto_slant" "rot13",
    "mirror",
    "katakana",
    "cards",
    "eftichess",
    "heart_left",
    "trashman",
    "morse",
    "eftipiti",
    "smtengwar",
    "e__fist_",
    "mike",
    "bear",
    "hills___",
    "rotated",
    "wow",
    "eftipiti",
    "relief2",
    "mshebrew210",
    "kik_star",
    "puzzle",
    "p_skateb",
    "hypa_bal",
    "tomahawk",
    "timesofl",
    "moscow",
    "cola",
    "baz__bil",
    "stencil1",
    "battlesh",
    "tsn_base",
    "kgames_i",
    "binary",
    "greek",
    "mnemonic",
    "panther_",
    "b1ff",
    "c_consen",
    "horizontal_right",
    "dwhistled",
    "hex",
    "flipped",
    "high_noo",
    "patorjk-hex",
    "amc_3_liv1",
    "gauntlet",
    "cybersmall",
    "octal",
    "js_cursive",
    "battle_s",
    "deep_str",
    "rally_s2",
    "convoy__",
    "atc_gran",
    "grand_pr",
    "ivrit",
    "rammstein",
    "horizontal_left",
    "eftiwall",
    "decimal",
    "goofy",
    "rot13",
    "konto_slant",
    "subteran",
    "rally_sp",
    "charset_",
]
ALL_FONTS = pyfiglet.FigletFont.getFonts()
OK_FONTS = list(filter(lambda x: x not in BAD_FONTS, ALL_FONTS))

DATASET_NAME = "figlet_font"


@dataclass
class FigletFontConfig:
    """Configuration for FigletFont task generation"""

    static_word: Optional[str] = None
    static_font: Optional[str] = None
    min_word_len: int = 3
    max_word_len: int = 7
    space_letters: bool = True
    seed: Optional[int] = None
    size: int = 500

    def validate(self):
        assert self.min_word_len > 0, "min_word_len must be greater than 0"
        assert self.min_word_len <= self.max_word_len, "min_word_len must be less than or equal to max_word_len"
        if self.static_word:
            assert len(self.static_word) > 0, "static_word must have at least one character"
        if self.static_font:
            assert len(self.static_font) > 0, "static_font must have at least one character"
            assert self.static_font in OK_FONTS, f"static_font must be one of {OK_FONTS}"


class FigletFontDataset(ProceduralDataset):
    """Generates FigletFont tasks"""

    def __init__(self, config: FigletFontConfig):
        with get_data_file_path("anagrams.jsonl").open() as f:
            self.words = [
                word
                for line in f
                for word in json.loads(line)["words"]
                if config.min_word_len <= len(word) <= config.max_word_len
            ]
            assert len(self.words) > 0, "No words found in the dataset with the specified length range"

        self._prompt_templates = [
            "What word does this say?\n\n{figlet_render}",
            "Please read the following figlet font:\n\n{figlet_render}",
        ]
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single FigletFont task

        Returns:
            dict with keys:
                - question: str, the task description with figlet string
                - answer: str, the figlet encoded word
                - metadata: dict with generation parameters
        """
        rng = Random(self.seed + idx)

        word = self.config.static_word if self.config.static_word is not None else rng.choice(self.words).upper()
        if self.config.space_letters:
            render_word = " ".join(word)
        else:
            render_word = word

        chosen_font = self.config.static_font if self.config.static_font is not None else rng.choice(OK_FONTS)
        figlet_render = pyfiglet.figlet_format(render_word, font=chosen_font)

        return {
            "question": rng.choice(self._prompt_templates).format(figlet_render=figlet_render),
            "answer": word,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "font": chosen_font,
                "space_letters": self.config.space_letters,
                "difficulty": {
                    "word_len": (self.config.min_word_len, self.config.max_word_len),
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided solves the figlet task.

        The function awards 1.0 for a correct answer and 0.1 points for each correct letter in the correct position,
        with a maximum possible score of 1.0.

        Args:
            answer (Optional[str]): The user's answer.
            entry (dict[str, Any]): The original dataset entry containing the correct answer.

        Returns:
            float: The computed score between 0.0 and 1.0.
        """

        correct_word = entry["answer"]
        if not isinstance(answer, str):
            return 0.0  # No answer given

        # Normalize case
        answer = answer.replace(" ", "").strip().lower()
        correct_word = correct_word.strip().lower()

        if answer == correct_word:
            return 1.0  # Correct!

        # Calculate similarity
        correct_count = sum(1 for a, b in zip(answer, correct_word) if a == b)
        max_length = max(len(correct_word), len(answer))

        # Compute a partial score
        score = min(correct_count * 0.1, 1.0)

        return score


class FigletFontCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(FigletFontCurriculum.__name__, FigletFontConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="word_len",
                levels=[3, 5, 10, 15, 20, 30],
                default_level=0,
                description="The length of the word to be displayed",
                lower_field_name="min_word_len",
                upper_field_name="max_word_len",
                ensure_interval=True,
            ),
        )


# Register the dataset
register_dataset(DATASET_NAME, FigletFontDataset, FigletFontConfig, FigletFontCurriculum)



import gzip
import json
from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import zss

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..data import get_data_file_path
from ..factory import ProceduralDataset, register_dataset

OUTPUT_PREDICTION_PROMPT_TEMPLATE = """
You are given a question that requires some input and output variables as follows:

{0}

The input and output requirements are as follows:

{1}

Given the following input:

{2}

Can you predict the output without writing any code? Please think and then provide the exact output in the form of a JSON value (object, array, number or string, or one of the following three literal names: false null true) as your final answer. The keys and values of the object should strictly match the output requirement as specified.

Tip: Here is a reference code snippet for this question. You can refer to this code to guide your reasoning but not copy spans of code directly.

{3}
"""

INPUT_PREDICTION_PROMPT_TEMPLATE = """
You are given a question that requires some input and output variables as follows:

{0}

The input and output requirements are as follows:

{1}

Given the following output:

{2}

Can you predict a feasible input without writing any code? Please reason and put your final answer in the form of a JSON value (object, array, number or string, or one of the following three literal names: false null true), even if the there is only one input variable, with keys strictly matching the input variables' names as specified.

Tip: Here is a reference code snippet for this question. You can refer to this code to guide your reasoning but not copy spans of code directly.

{3}
"""

DATASET_NAME = "codeio"


@dataclass
class CodeIOConfig:
    """Configuration for CodeI/O reasoning task generation"""

    seed: Optional[int] = None
    size: int = 500
    input_prediction_probability: float = 0.5
    difficulty: Optional[int] = None

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert 0.0 <= self.input_prediction_probability <= 1.0, "input_prediction_probability must be in [0, 1]"
        if self.difficulty is not None:
            assert 1 <= self.difficulty <= 10, "difficulty must be in [1, 10]"


class CodeIODataset(ProceduralDataset):
    """
    Exercise some caution when using this dataset, as it involves executing arbitrary code snippets.
    These code snippets are transformed by an LLM from raw code files which have been curated from high-quality sources.
    However, there is still a risk that the LLM could have introduced code with bad effects.
    """

    _jsonl_data: Optional[list] = None

    def __init__(self, config: CodeIOConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

        self._data_path = get_data_file_path("codeio.jsonl.gz")

        with gzip.open(self._data_path, "rt", encoding="utf-8") as f:
            data = [json.loads(line) for line in f]
            if self.config.difficulty is not None:
                data = [entry for entry in data if entry.get("difficulty", -1) == self.config.difficulty]
            assert len(data) > 0, "No data found for the specified difficulty level"
            CodeIODataset._jsonl_data = data

    def _generate_io_pair(self, main_code: str, input_generator_code: str, rng: Random, max_retries: int = 1):
        local_vars = {"Random": Random}

        full_code = f"{main_code}\n\n{input_generator_code}"
        try:
            exec(full_code, local_vars, local_vars)
        except Exception as e:
            print(f"Error executing code:\n{full_code}")
            print(f"---------------------\nException: {e}\n---------------------")
            return {}, {}

        for _ in range(max_retries):
            try:
                inputs = local_vars["generate_inputs"](rng)
                outputs = local_vars["main_solution"](**inputs)
            except Exception as e:
                # Retry
                print(f"Error generating I/O pair: {e}")
                continue
            return inputs, outputs
        return {}, {}

    def __getitem__(self, idx: int) -> dict:
        """Generate a single CodeI/O reasoning task"""
        rng = Random(self.seed + idx)

        json_data = rng.choice(CodeIODataset._jsonl_data)

        query = json_data["task_description"]
        parameters = json_data["input_output_spec"]
        reference_code = json_data["code_sample"]
        input_generator_code = json_data["input_generator"]

        input_data, output_data = self._generate_io_pair(reference_code, input_generator_code, rng)

        if rng.random() < self.config.input_prediction_probability:
            question = OUTPUT_PREDICTION_PROMPT_TEMPLATE.format(query, parameters, input_data, reference_code)
            solution = json.dumps(output_data)
        else:
            question = INPUT_PREDICTION_PROMPT_TEMPLATE.format(query, parameters, output_data, reference_code)
            solution = json.dumps(input_data)

        return {
            "question": question,
            "answer": solution,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "input_data": input_data,
                "output_data": output_data,
                "difficulty": {"difficulty": self.config.difficulty},
            },
        }

    def _json_to_tree(self, data, label="root"):
        """Recursively convert a JSON dictionary to a ZSS tree."""
        if isinstance(data, dict):
            node = zss.Node(label)
            for key, value in sorted(data.items()):
                node.addkid(self._json_to_tree(value, key))
            return node
        elif isinstance(data, list):
            node = zss.Node(label)
            for idx, item in enumerate(data):
                node.addkid(self._json_to_tree(item, f"item_{idx}"))
            return node
        else:
            return zss.Node(f"{label}:{data}")

    def _compute_json_similarity(self, json1, json2):
        """Compute a similarity score in [0, 1] between two JSON dictionaries using tree edit distance."""
        tree1 = self._json_to_tree(json1)
        tree2 = self._json_to_tree(json2)

        def _str_edit_distance(str1, str2):
            """Compute Levenshtein edit distance between two strings."""
            m, n = len(str1), len(str2)
            prev = list(range(n + 1))
            curr = [0] * (n + 1)
            for i in range(1, m + 1):
                curr[0] = i
                for j in range(1, n + 1):
                    if str1[i - 1] == str2[j - 1]:
                        curr[j] = prev[j - 1]
                    else:
                        curr[j] = 1 + min(prev[j], curr[j - 1], prev[j - 1])
                prev, curr = curr, prev
            return prev[n]

        def _tree_node_edit_distance(text1: str, text2: str):
            """Compute edit distance between two tree nodes based on their types."""
            if ":" not in text1 or ":" not in text2:
                return _str_edit_distance(text1, text2)

            key1, value1 = text1.split(":", 1)
            key2, value2 = text2.split(":", 1)

            key_dist = _str_edit_distance(key1, key2) if key1 != key2 else 0
            value_dist = _str_edit_distance(value1, value2) if value1 != value2 else 0

            if value1 != value2:
                # Numeric, allowing decimals
                if value1.replace(".", "").isnumeric() and value2.replace(".", "").isnumeric():
                    try:
                        # TODO: Consider a more sophisticated distance metric for numeric values?
                        abs1, abs2 = abs(float(value1)), abs(float(value2))
                        divisor = max(min(abs1, abs2), 10e-5)
                        value_dist = max(len(value1), len(value2)) * abs((abs1 - abs2) / divisor)
                    except ValueError:
                        # Fall back on string edit distance
                        pass
                elif value1.isnumeric() or value2.isnumeric():
                    # Penalise severely if the answer is numeric when it shouldn't be, or vice versa
                    value_dist += max(len(text1), len(text2))

            return key_dist + value_dist

        edit_distance = zss.simple_distance(tree1, tree2, label_dist=_tree_node_edit_distance)
        max_size = max(len(json.dumps(json1)), len(json.dumps(json2)))

        similarity_score = 1 - (edit_distance / (0.2 * max_size))
        return max(0, similarity_score)

    def _score_answer_json(self, answer_json: dict, oracle_json: dict, max_score: float) -> float:
        """If the answer is valid JSON, compute a similarity score between the answer and the oracle JSON."""
        if answer_json == oracle_json:
            return max_score
        else:
            similarity = self._compute_json_similarity(answer_json, oracle_json)
            # 0.01 minimum reward, since it produced a valid JSON output
            return max(similarity * max_score, 0.01)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        oracle_answer = entry["answer"].strip()
        reward = 0.0
        if answer is not None and len(answer) > 0:
            answer = answer.strip()
            if answer == oracle_answer:
                reward = 1.0
            elif "{" in answer and "}" in answer:
                # Check if the answer contains a correct format JSON object somewhere
                # But penalise for length & accuracy
                ans_first_open, ans_last_close = answer.index("{"), answer.rindex("}")
                extra_chars = len(answer[:ans_first_open]) + len(answer[ans_last_close + 1 :])

                # 0.5 is arbitrary here, but the answers are very short so it seems harsh to penalize too much
                # e.g. if oracle is {"steps": "3"} and answer is "The correct answer is: {"steps": "3"}"
                max_score = max(len(oracle_answer) / (len(oracle_answer) + 0.5 * extra_chars), 0.2)

                try:
                    answer_dict = json.loads(answer[ans_first_open : ans_last_close + 1])
                    oracle_dict = json.loads(oracle_answer)
                    return self._score_answer_json(answer_dict, oracle_dict, max_score)
                except json.JSONDecodeError:
                    if oracle_answer in answer:
                        reward = len(oracle_answer) / len(answer)
                    else:
                        reward = 0.00
            else:
                reward = 0.00

        return reward


class CodeIOCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(CodeIOCurriculum.__name__, CodeIOConfig)

        self._define_attributes(
            ScalarAttributeDefinition(
                name="difficulty",
                field_name="difficulty",
                levels=[6, 7, 8, 9],
                description="Difficulty level of the task",
            ),
        )


# Register the dataset
register_dataset(DATASET_NAME, CodeIODataset, CodeIOConfig, CodeIOCurriculum)



"""Base class for procedural dataset generators"""

from abc import ABC, abstractmethod
from collections.abc import Iterable, Sized
from copy import deepcopy
from random import Random
from typing import Any, Iterator, Optional, Type, TypeVar


class ProceduralDataset(ABC, Sized, Iterable[dict[str, Any]]):
    """Abstract base class for procedural dataset generators"""

    def __init__(self, config: Any, seed: Optional[int] = None, size: int = 500):
        """Initialize the dataset with config, optional seed and size"""
        if hasattr(config, "validate") and callable(config.validate):
            config.validate()

        self.config = config
        self.size = size
        self.seed = seed if seed is not None else Random().randint(0, 2**32)

    @property
    def category(self) -> str:
        """Extract category from the module name."""
        module_name = self.__class__.__module__
        parts = module_name.split(".")
        if len(parts) >= 3:
            return parts[1]  # reasoning_gym.{category}.dataset_name
        return "other"

    def __len__(self) -> int:
        """Return the virtual size of the dataset"""
        return self.size

    def __iter__(self):
        """Make the dataset iterable"""
        self._current_idx = 0
        return self

    def __next__(self) -> dict[str, Any]:
        """Get next item in iteration"""
        if self._current_idx >= self.size:
            raise StopIteration
        item = self[self._current_idx]
        self._current_idx += 1
        return item

    @abstractmethod
    def __getitem__(self, idx: int) -> dict[str, Any]:
        """Generate a single dataset item

        Args:
            idx: Index of the item to generate

        Returns:
            dict containing at least:
                - question: str
                - answer: str
                - metadata: dict
        """
        raise NotImplementedError

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Overwrite this method in derived classes if a single oracle answer is not available."""
        oracle_answer = entry["answer"]
        reward = 0.0
        if isinstance(answer, str) and len(answer) > 0:
            if answer == oracle_answer:
                reward = 1.0
            elif oracle_answer in answer:
                reward = len(oracle_answer) / len(answer)
        return reward


T = TypeVar("T", bound="ProceduralDataset")


class ReseedingDataset(Iterable[dict[str, Any]]):
    """Wrapper that makes any ProceduralDataset infinite by reseeding when reaching the end"""

    def __init__(self, dataset: T, chunk_size: int = 500):
        """Initialize with dataset instance and chunk size

        Args:
            dataset: The ProceduralDataset instance to wrap
            chunk_size: Size of each generated chunk before reseeding
        """
        self.dataset = dataset
        self.dataset_cls: Type[T] = type(dataset)
        self.chunk_size = chunk_size

        # Start with chunk 0
        self._current_chunk = 0
        self._current_dataset = self._create_chunk(0)
        self._current_idx = 0

    def _create_chunk(self, chunk_num: int) -> T:
        """Create a new dataset chunk with unique seed"""
        # Create new config with modified seed
        new_config = deepcopy(self.dataset.config)
        if hasattr(new_config, "seed"):
            # Derive new seed from chunk number using dataset's seed, wrapping around at 2^32
            new_config.seed = (self.dataset.seed + chunk_num) % (2**32)

        # Create new dataset instance with chunk config
        return self.dataset_cls(new_config)

    def __iter__(self) -> Iterator[dict[str, Any]]:
        """Make the dataset iterable"""
        self._current_chunk = 0
        self._current_dataset = self._create_chunk(0)
        self._current_idx = 0
        return self

    def __next__(self) -> dict[str, Any]:
        """Get next item, creating new chunk if needed"""
        if self._current_idx >= self.chunk_size:
            # Move to next chunk
            self._current_chunk += 1
            self._current_dataset = self._create_chunk(self._current_chunk)
            self._current_idx = 0

        item = self._current_dataset[self._current_idx]
        self._current_idx += 1
        return item

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Forward scoring to the wrapped dataset's implementation"""
        return self.dataset.score_answer(answer, entry)



"""
Reasoning Gym - A library of procedural dataset generators for training reasoning models
"""

from . import (
    algebra,
    algorithmic,
    arc,
    arithmetic,
    code,
    cognition,
    data,
    games,
    geometry,
    graphs,
    induction,
    logic,
    probability,
)
from .factory import create_dataset, get_score_answer_fn, register_dataset

__version__ = "0.1.19"
__all__ = [
    "arc",
    "algebra",
    "algorithmic",
    "arithmetic",
    "code",
    "cognition",
    "data",
    "games",
    "geometry",
    "graphs",
    "logic",
    "induction",
    "probability",
    "create_dataset",
    "register_dataset",
    "get_score_answer_fn",
]



import random
import string
from dataclasses import dataclass
from typing import Any, Optional

from ..coaching import BaseCurriculum, RangeAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPALTE = """Your task is, given a list of letters, to form a valid palindrome.

A palindrome is a phrase that reads the same forwards and backwards.

If there are multiple possible answers, only respond with one of them. You must use all the letters provided.

Your output should be a single string, with no spaces or punctuation.

Now, form a valid palindrome using the following letters: {letters}
"""


DATASET_NAME = "palindrome_generation"


@dataclass
class PalindromeConfig:
    """
    Configuration for the palindrome task.

    - min_length: Minimum length of the palindrome.
    - max_length: Maximum length of the palindrome.
    - seed: Optional seed for reproducibility.
    - size: Number of palindrome samples in the virtual dataset.
    """

    min_length: int = 3
    max_length: int = 10
    seed: Optional[int] = None
    size: int = 50

    def validate(self) -> None:
        """Validate configuration parameters."""
        assert self.min_length >= 1, "min_length must be >= 1"
        assert self.max_length >= self.min_length, "max_length must be >= min_length"


class PalindromeDataset(ProceduralDataset):
    """
    Generates a set of letters that can be assembled into a palindrome.
    """

    def __init__(self, config: PalindromeConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """
        Generate a single palindrome task.

        Returns:
            dict with:
            - "question": Set of letters to form a palindrome.
            - "answer": A correct palindrome.
            - "metadata": Includes letter set and generated palindrome.
        """
        rng = random.Random(self.seed + idx)
        length = rng.randint(self.config.min_length, self.config.max_length)
        letters = self._generate_palindrome_letters(rng, length)
        scrambled_letters = rng.sample(letters, len(letters))  # Scramble the order
        palindrome = self._assemble_palindrome(letters)
        return {
            "question": QUESTION_TEMPALTE.format(letters=", ".join(scrambled_letters)),
            "answer": palindrome,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "letters": scrambled_letters,
                "generated_palindrome": palindrome,
                "length": length,
                "difficulty": {
                    "length": (self.config.min_length, self.config.max_length),
                },
            },
        }

    def _generate_palindrome_letters(self, rng: random.Random, length: int) -> list[str]:
        """Generate a set of letters that can form a palindrome."""
        half_length = length // 2
        letters = rng.choices(string.ascii_lowercase, k=half_length)
        if length % 2 == 1:
            middle_letter = rng.choice(string.ascii_lowercase)
            return letters + [middle_letter] + letters[::-1]
        return letters + letters[::-1]

    def _assemble_palindrome(self, letters: list[str]) -> str:
        """Return the palindrome string from the letter set."""
        return "".join(letters)

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided is a valid palindrome.
        The answer is expected to be a single string

        Expected behavior:
        - Correct answer (palindrome with only correct letters in the correct quantities) gives 1.0
        - An answer that is a palindrome, but not with the same letters as provided, gives 0.05
        - An answer that is a string, but not a palindrome gives 0.02
        - An empty string gives 0.0
        - None gives 0.0.
        """
        if answer is None or not isinstance(answer, str):
            return 0.0  # No answer given

        if answer == "":
            return 0.0

        metadata = entry["metadata"]
        answer = answer.strip().lower()
        expected_letters = metadata["letters"]

        # Check if the answer is a palindrome
        if answer != answer[::-1]:
            return 0.02

        # Check if answer contains the same letters as provided (ignoring order)
        if sorted(answer) != sorted(expected_letters):
            return 0.05

        return 1.0  # Correct solution


class PalindromeCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(PalindromeCurriculum.__name__, PalindromeConfig)

        # Define attributes
        self._define_attributes(
            RangeAttributeDefinition(
                name="length",
                levels=[10, 50, 100, 500],
                description="Length of the generated palindrome.",
                lower_field_name="min_length",
                upper_field_name="max_length",
                ensure_interval=True,
            )
        )


register_dataset(DATASET_NAME, PalindromeDataset, PalindromeConfig, PalindromeCurriculum)



"""Package containing data files used by reasoning_gym"""

from importlib import resources
from pathlib import Path


def get_data_file_path(filename: str) -> Path:
    """Get the path to a data file in the package.

    Args:
        filename: Name of the file in the data directory

    Returns:
        Path object pointing to the data file

    Example:
        >>> path = get_data_file_path("pg19362.txt")
        >>> with open(path, encoding="utf-8") as f:
        ...     content = f.read()
    """
    return resources.files("reasoning_gym.data").joinpath(filename)


def read_data_file(filename: str) -> str:
    """Read the contents of a data file in the package.

    Args:
        filename: Name of the file in the data directory

    Returns:
        String contents of the file

    Example:
        >>> content = read_data_file("pg19362.txt")
    """
    return resources.files("reasoning_gym.data").joinpath(filename).read_text(encoding="utf-8")


__all__ = ["get_data_file_path", "read_data_file"]



class Token:

    INT = "INT"
    VOID = "VOID"
    TRUE = "TRUE"
    FALSE = "FALSE"
    AND = "AND"
    OR = "OR"
    NOT = "NOT"
    RETURN = "RETURN"
    IF = "IF"
    ELSE = "ELSE"
    WHILE = "WHILE"
    FOR = "FOR"
    DO = "DO"
    BREAK = "BREAK"
    CONTINUE = "CONTINUE"
    SWITCH = "SWITCH"
    CASE = "CASE"
    DEFAULT = "DEFAULT"
    COLON = "COLON"
    SEMICOLON = "SEMICOLON"
    COMMA = "COMMA"

    LPAREN = "LPAREN"
    RPAREN = "RPAREN"
    LBRACE = "LBRACE"
    RBRACE = "RBRACE"
    LBRACK = "LBRACK"
    RBRACK = "RBRACK"

    ASSIGN = "ASSIGN"
    TERNARY = "TERNARY"
    RELOP = "RELOP"
    BINOP = "BINOP"
    INCREMENT = "INCREMENT"
    DECREMENT = "DECREMENT"
    UNARY_MULTIPLICATIVE = "UNARY_MULTIPLICATIVE"

    BITWISE_SHIFT = "BITWISE_SHIFT"
    BITWISE_NOT = "BITWISE_NOT"
    BITWISE_AND = "BITWISE_AND"
    BITWISE_OR = "BITWISE_OR"
    BITWISE_XOR = "BITWISE_XOR"

    WHITESPACE = "WHITESPACE"
    ID = "ID"
    NUM = "NUM"
    STRING = "STRING"
    CHAR = "CHAR"

    PRINT = "PRINT"
    COMMENT = "COMMENT"
    UNIDENTIFIED = "UNIDENTIFIED"

    def __init__(self, type, line, column, data=None, original_tokens=None):
        self.type = type
        self.line = line
        self.column = column
        self.data = data
        self.original_tokens = original_tokens

    def __str__(self):
        result = self.type
        if self.data:
            result += " " + self.data
        result += " (line %s column %s)" % (self.line, self.column)
        if self.original_tokens:
            result += " (original tokens: " + ", ".join([str(t) for t in self.original_tokens]) + ")"
        return result



"""List functions generators"""

from dataclasses import dataclass
from random import Random
from typing import Any, Callable, Optional

from reasoning_gym.factory import ProceduralDataset, register_dataset

DATASET_NAME = "list_functions"


@dataclass
class ListFunctionsDatasetConfig:
    """Configuration for List function generators."""

    seed: Optional[int] = None
    size: int = 500

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.size > 0, "size must be positive"


tasks = list(range(17))


class ListFunctionsDataset(ProceduralDataset):

    def __init__(self, config: ListFunctionsDatasetConfig):
        super().__init__(config, config.seed, config.size)
        self._generators: dict[int, Callable[[Random, float], dict[str, Any]]] = None  # initially None, lazy loading
        self.task_indices = Random(self.seed).choices(tasks, k=self.size)
        self.prompt_template = """You are an expert at inductive reasoning. Generate an output corresponding to the given input.
The output is generated by applying the same rule that maps input to output for the examples provided. Your answer should be a list of element/elements
Examples:
{examples}

Input: {input}
Output:
"""

    @property
    def generators(self) -> dict[int, Callable[[Random, float], dict[str, Any]]]:
        """Lazy load generators only when first accessed"""
        if self._generators is None:
            self._generators = self._load_generators()
        return self._generators

    def _load_generators(self):
        """
        Generates mapper from task identifiers (keys) to example generator functions
        """
        from . import generators

        def strip_prefix(s: str, prefix: str) -> str:
            return s[len(prefix) :]

        prefix = "generate_"
        gs = {}
        for n in dir(generators):
            if n.startswith(prefix):
                gs[int(strip_prefix(n, prefix))] = getattr(generators, n)
        return gs

    def __getitem__(self, idx: int) -> dict:
        """Generate a single induction-based list function dataset"""
        rng = Random(self.seed + idx)
        generator_idx = self.task_indices[idx]
        generator = self.generators[generator_idx]
        examples = generator(rng)
        entry = examples.popitem()
        input = entry[0]
        output = entry[1]
        formatted_examples = ""
        for index, key in enumerate(examples):
            formatted_examples += f"""Input {index + 1}: {key}
Output {index + 1}: {examples[key]}
"""
        question = self.prompt_template.format(examples=formatted_examples, input=input)
        return {
            "question": question,
            "answer": output,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
            },
        }


register_dataset(DATASET_NAME, ListFunctionsDataset, ListFunctionsDatasetConfig)



import random
import string
from dataclasses import dataclass, field
from typing import Optional

from sympy import Symbol

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "simple_equations"


@dataclass
class SimpleEquationsConfig:
    """Configuration for simple equation task generation"""

    min_terms: int = 2  # Minimum number of terms in expression
    max_terms: int = 4  # Maximum number of terms
    min_value: int = 1  # Minimum value for constants
    max_value: int = 100  # Maximum value for constants
    operators: tuple = ("+", "-", "*")  # Allowed operators
    operators_weights: list[float] = field(default_factory=lambda: [0.4, 0.4, 0.2])  # Weights for each operator
    seed: Optional[int] = None
    size: int = 500

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.min_terms > 0, "min_terms must be positive"
        assert self.max_terms >= self.min_terms, "max_terms must be >= min_terms"
        assert self.min_value > 0, "min_value must be positive"
        assert self.max_value >= self.min_value, "max_value must be >= min_value"
        assert len(self.operators) > 0, "must specify at least one operator"
        assert all(op in ("+", "-", "*") for op in self.operators), "invalid operator specified"
        assert round(sum(self.operators_weights), 1) == 1.0, "operators_weights must sum to 1.0"


class SimpleEquationsDataset(ProceduralDataset):
    """Generates simple equations with one variable to solve"""

    def __init__(self, config: SimpleEquationsConfig):
        self._prompt_templates = [
            "Find the value of {variable} in the equation: {equation}",
            "Solve for {variable}: {equation}",
            "Determine the value of {variable} that satisfies: {equation}",
        ]
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single equation task

        Returns:
            dict with keys:
                - question: str, the equation to solve (e.g. "3 * x = 12")
                - answer: str, the solution value (e.g. "4")
                - metadata: dict with generation parameters
        """
        rng = random.Random(self.seed + idx)

        # Get variable and generate equation
        variable = self._get_variable(rng)
        equation, solution = self._generate_equation(rng, variable)

        return {
            "question": rng.choice(self._prompt_templates).format(variable=variable, equation=equation),
            "answer": str(solution),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "equation": equation,
                "variable": variable,
                "difficulty": {
                    "min_terms": self.config.min_terms,
                    "max_terms": self.config.max_terms,
                    "min_value": self.config.min_value,
                    "max_value": self.config.max_value,
                    "operators_weights": self.config.operators_weights,
                },
            },
        }

    def _get_variable(self, rng: random.Random) -> str:
        """Get a random lowercase variable name"""
        return rng.choice(string.ascii_lowercase)

    def _generate_equation(self, rng: random.Random, variable: str) -> tuple[str, int]:
        """Generate an equation and its solution

        Args:
            rng: Random number generator
            variable: Variable symbol to use in equation

        Returns:
            Tuple of (equation string, solution integer)
        """
        x = Symbol(variable)

        # Generate terms for left side
        num_terms = rng.randint(self.config.min_terms, self.config.max_terms)
        terms = []

        # Generate all constant terms first
        for _ in range(num_terms):
            value = rng.randint(self.config.min_value, self.config.max_value)
            terms.append(value)

        # Replace one random term with the variable term
        var_pos = rng.randint(0, num_terms - 1)
        coef = rng.randint(self.config.min_value, self.config.max_value)
        if "*" in self.config.operators:
            terms[var_pos] = coef * x
        else:
            terms[var_pos] = x

        # Apply operators between terms
        expr = terms[0]
        for i in range(1, num_terms):
            op = rng.choices(self.config.operators, weights=self.config.operators_weights, k=1)[0]
            if op == "+":
                expr = expr + terms[i]
            elif op == "-":
                expr = expr - terms[i]
            else:  # '*'
                expr = expr * terms[i]

        left_side = expr
        solution_value = rng.randint(self.config.min_value, self.config.max_value)
        right_side = left_side.subs(x, solution_value)
        return f"{left_side} = {right_side}", solution_value


class SimpleEquationsCurriculum(BaseCurriculum):
    """Curriculum for simple equations task"""

    def __init__(self):
        super().__init__(SimpleEquationsCurriculum.__name__, SimpleEquationsConfig)

        # Define attributes
        self._define_attributes(
            ScalarAttributeDefinition(
                name="min_terms",
                field_name="min_terms",
                levels=[2, 3, 4, 5],
                description="Minimum number of terms in simple equations",
            ),
            ScalarAttributeDefinition(
                name="max_terms",
                field_name="max_terms",
                levels=[5, 10, 15, 20],
                description="Maximum number of terms in simple equations",
            ),
            ScalarAttributeDefinition(
                name="min_value",
                field_name="min_value",
                levels=[1, 10, 100, 1000],
                description="Minimum value for constants in simple equations",
            ),
            ScalarAttributeDefinition(
                name="max_value",
                field_name="max_value",
                levels=[100, 10000, 1000000, 100000000],
                description="Maximum value for constants in simple equations",
            ),
            ScalarAttributeDefinition(
                name="operators_weights",
                field_name="operators_weights",
                levels=[[0.4, 0.4, 0.2], [0.35, 0.35, 0.3], [0.3, 0.3, 0.4], [0.2, 0.2, 0.6]],
                description="Weights for each operator in simple equations",
            ),
        )


register_dataset(DATASET_NAME, SimpleEquationsDataset, SimpleEquationsConfig, SimpleEquationsCurriculum)



from dataclasses import dataclass
from random import Random
from typing import Any, Optional

import bfi

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..data.wordle_words import wordle_words
from ..factory import ProceduralDataset, register_dataset
from .contrib.bfit.Compiler import Compiler, Minify

DATASET_NAME = "bf"


@dataclass
class BFConfig:
    """Configuration for BF task generation"""

    seed: Optional[int] = None
    size: int = 500
    difficulty: int = 1

    def validate(self) -> None:
        """Validate configuration parameters"""
        assert self.difficulty > 0, "difficulty must be greater than 0"
        assert self.difficulty < 4, "difficulty must be less than 4"


class BFDataset(ProceduralDataset):
    """Generates BF tasks"""

    def __init__(self, config: BFConfig):
        self._prompt_templates = [
            "This is a BF (Brainf*ck) computer program. What is the output?\n\n{bf_program}\n\nRespond only with the exact output of the program.",
            "Consider the following BF (Brainf*ck) code. What would it output?\n\n{bf_program}\n\nProvide only the exact output of the code.",
        ]
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single BF task

        Returns:
            dict with keys:
                - question: str, the task description with BF program
                - answer: str, the result of this BF program BFI execution
                - metadata: dict with generation parameters
        """
        rng = Random(self.seed + idx)

        bfit_code = self.generate_bfit_code(self.config.difficulty, rng)
        bf_program = self.compile_bfit_code_to_bf(bfit_code)
        result = bfi.interpret(bf_program, buffer_output=True)

        return {
            "question": rng.choice(self._prompt_templates).format(bf_program=bf_program),
            "answer": result,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "bfit_code": bfit_code,
                "bf_program": bf_program,
                "difficulty": {"difficulty": self.config.difficulty},
            },
        }

    def generate_bfit_code(self, difficulty, rng: Random) -> str:

        if difficulty == 1:
            word = rng.choice(wordle_words)
            bfit_template = f"""
int main() {{
    print("{word}");
}}
"""
        elif difficulty == 2:
            x = rng.randint(1, 4)
            y = rng.randint(1, 5)
            target = x * y * rng.randint(1, 9) + rng.randint(1, 9)
            bfit_template = f"""
int main() {{
    int acc = 0;
    int target = {target};
    int x = {x};
    int y = {y};
    while (acc < target) {{
        acc = acc + x;
        acc = acc + y;
    }}
    printint(acc);
}}
"""
        elif difficulty == 3:
            x = rng.randint(1, 7)
            y = rng.randint(1, 9)
            target = x * y * rng.randint(1, 9) + rng.randint(1, 9) + 50
            conditional = target - rng.randint(1, 40)
            bfit_template = f"""
int main() {{
    int acc = 0;
    int target = {target};
    int x = {x};
    int y = {y};
    while (acc < target) {{
        acc = acc + x;
        if (acc > {conditional}) {{
            acc = acc + y;
        }}
    }}
    printint(acc);
}}
"""
        rendered_bfit = bfit_template
        return rendered_bfit

    def compile_bfit_code_to_bf(self, bfit: str) -> str:
        bf = Compiler.compile(bfit, optimize_code=True)
        # bf = Minify.minify(bf) # Is this necessary?
        return bf


class BFCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(BFCurriculum.__name__, BFConfig)

        # Define attributes
        self._define_attributes(
            ScalarAttributeDefinition(
                name="difficulty",
                field_name="difficulty",
                levels=[1, 2, 3],
                description="Difficulty level",
            )
        )


# Register the dataset
register_dataset(DATASET_NAME, BFDataset, BFConfig, BFCurriculum)



"""
Arithmetic tasks for training reasoning capabilities:
"""

from .acre.acre import ACREDataset, ACREDatasetConfig
from .list_functions import ListFunctionsDataset, ListFunctionsDatasetConfig

__all__ = ["ListFunctionsDataset", "ListFunctionsDatasetConfig", "ACREDataset", "ACREDatasetConfig"]



from dataclasses import dataclass
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

QUESTION_TEMPLATE = """Your task is to count how many rectangles are present in an ASCII grid.

Single rectangles are outlined with a '#', overlapping rectangles (max 2) are shown with '█'.

Your output should be a single number, representing the total count of rectangles.

Now, it's your turn. How many rectangles do you see in the grid below?
{puzzle}
"""

DATASET_NAME = "rectangle_count"
CONST_TERM = 0.8
D = 5


def draw_rectangles_with_overlap(n, width, height, rng):
    # Create a grid that holds a count of how many times a cell is drawn.
    grid = [[0 for _ in range(width)] for _ in range(height)]
    rectangles = []

    max_attempts = 100000  # Prevent infinite loops in case of a crowded grid
    attempts = 0

    while len(rectangles) < n and attempts < max_attempts:
        attempts += 1
        # Ensure minimum width and height of 3.
        # For a rectangle to be at least 3 cells wide, right must be at least left + 2.
        # Similarly, bottom must be at least top + 2.
        left = rng.randint(0, width - 3)
        right = rng.randint(left + 2, width - 1)
        top = rng.randint(0, height - 3)
        bottom = rng.randint(top + 2, height - 1)

        # Prepare a list of all the cells that would be updated.
        cells_to_update = []

        # Top edge:
        for col in range(left, right + 1):
            cells_to_update.append((top, col))
        # Bottom edge:
        for col in range(left, right + 1):
            cells_to_update.append((bottom, col))
        # Left edge (excluding corners already drawn):
        for row in range(top + 1, bottom):
            cells_to_update.append((row, left))
        # Right edge (excluding corners already drawn):
        for row in range(top + 1, bottom):
            cells_to_update.append((row, right))

        # Check if drawing this rectangle would cause any cell to exceed a count of 2.
        conflict = False
        for r, c in cells_to_update:
            if grid[r][c] >= 2:
                conflict = True
                break
        if conflict:
            continue  # Skip this rectangle candidate

        # No conflict: update the grid counts.
        for r, c in cells_to_update:
            grid[r][c] += 1

        # Save the rectangle (stored as (left, right, top, bottom)).
        rectangles.append((left, right, top, bottom))

    if len(rectangles) < n:
        print(f"Only placed {len(rectangles)} rectangles after {attempts} attempts.")

    # Print the grid.
    # Use ' ' for an untouched cell, '#' for a single hit, and '█' for exactly two hits.
    lines = ""
    for row in grid:
        line = "".join(" " if count == 0 else ("#" if count == 1 else "█") for count in row)
        lines = lines + line + "\n"
    return lines, len(rectangles)


@dataclass
class RectangleCountConfig:
    """Configuration for RectangleCount puzzle generation"""

    max_rectangles: int = 10
    width: int = 80
    height: int = 80
    seed: Optional[int] = None
    size: int = 500

    def validate(self):
        """Validate configuration parameters"""
        assert self.width >= 10, "width must be gte 10"
        assert self.height >= 10, "height must be gte 10"


class RectangleCountDataset(ProceduralDataset):
    """Generates ASCII rectangle counting puzzles with configurable parameters"""

    def __init__(self, config: RectangleCountConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single RectangleCount task

        Returns:
            dict with keys:
                - question: str, the task description
                - answer: str, a solution string
                - metadata: dict with generation parameters
        """
        rng = Random(self.seed + idx)

        target = rng.randint(1, self.config.max_rectangles)
        puzzle, answer = draw_rectangles_with_overlap(target, self.config.width, self.config.height, rng)

        return {
            "question": QUESTION_TEMPLATE.format(puzzle=puzzle),
            "answer": str(answer),
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "puzzle": puzzle,
                "solution": answer,
                "num_rectangles": target,
                "difficulty": {
                    "max_rectangles": self.config.max_rectangles,
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided solves the RectangleCount task,
        awarding partial credit if the guess is close.

        Returns:
            float: A score between 0.0 and 1.0.
        """
        correct_str = entry["answer"].lower().replace("\n", "")

        try:
            correct_val = int(correct_str)
            user_val = int(answer.strip())
        except (ValueError, TypeError, AttributeError):
            return 0.0
        distance = abs(user_val - correct_val)

        if distance == 0:
            return 1.0
        if distance >= D:
            return 0.0

        score = 1.0 - (distance / float(D))
        score = CONST_TERM * score
        return max(0.0, score)


class RectangleCountCurriculum(BaseCurriculum):
    def __init__(self):
        super().__init__(RectangleCountCurriculum.__name__, RectangleCountConfig)

        # Define attributes
        self._define_attributes(
            ScalarAttributeDefinition(
                name="max_rectangles",
                levels=[5, 10, 15, 20, 25],
                description="Number of rectangles in the grid",
                field_name="max_rectangles",
            ),
        )


register_dataset(DATASET_NAME, RectangleCountDataset, RectangleCountConfig, RectangleCountCurriculum)



from collections import defaultdict
from heapq import heappop, heappush

import numpy as np

from reasoning_gym.games.contrib.sokoban.src.utils import (
    can_move,
    dijkstra_sum,
    get_state,
    is_deadlock,
    is_solved,
    manhattan_sum,
)


def astar(matrix, player_pos, debug: bool = False, heuristic: str = "manhattan", max_depth: int = 100):
    # print(f'A* - {heuristic.title()} Heuristic')
    heur = "[A*]" if heuristic == "manhattan" else "[Dijkstra]"
    shape = matrix.shape
    initial_state = get_state(matrix)
    initial_cost = curr_depth = 0
    if heuristic == "manhattan":
        curr_cost = manhattan_sum(initial_state, player_pos, shape)
    else:
        distances = defaultdict(lambda: [])
        curr_cost = dijkstra_sum(initial_state, player_pos, shape, distances)
    seen = {None}
    heap = []
    heappush(heap, (initial_cost, curr_cost, initial_state, player_pos, curr_depth, ""))
    moves = [(1, 0), (-1, 0), (0, -1), (0, 1)]
    direction = {
        (1, 0): "D",
        (-1, 0): "U",
        (0, -1): "L",
        (0, 1): "R",
    }
    while heap:
        _, curr_cost, state, pos, depth, path = heappop(heap)
        seen.add(state)
        for move in moves:
            new_state, move_cost = can_move(state, shape, pos, move)
            deadlock = is_deadlock(new_state, shape)
            if new_state in seen or deadlock:
                continue
            new_pos = pos[0] + move[0], pos[1] + move[1]
            if heuristic == "manhattan":
                new_cost = manhattan_sum(new_state, new_pos, shape)
            else:
                new_cost = dijkstra_sum(new_state, new_pos, shape, distances)
            if new_cost == float("inf"):
                continue
            heappush(
                heap,
                (
                    move_cost + curr_cost,
                    new_cost,
                    new_state,
                    new_pos,
                    depth + 1,
                    path + direction[move],
                ),
            )
            if is_solved(new_state):
                # print(f'{heur} Solution found!\n\n{path + direction[move]}\nDepth {depth + 1}\n')
                if debug:
                    print(f"{heur} Solution Found!\n{path + direction[move]}", 20)
                return (path + direction[move], depth + 1)
            if debug:
                print(f"{heur} Solution Depth: {depth + 1}\n{path + direction[move]}", 20)

        if depth > max_depth:
            break

    if debug:
        print(f"{heur} Solution Not Found!\nDepth {depth + 1}", 20)

    return (None, -1 if not heap else depth + 1)


def solve_astar(puzzle, visualizer: bool = False, heuristic: str = "manhattan", max_depth: int = 100):
    matrix = puzzle
    where = np.where((matrix == "*") | (matrix == "%"))
    player_pos = where[0][0], where[1][0]
    return astar(matrix, player_pos, debug=visualizer, heuristic=heuristic, max_depth=max_depth)



from dataclasses import dataclass
from functools import reduce
from math import gcd
from random import Random
from typing import Any, Optional

from ..coaching import BaseCurriculum, ScalarAttributeDefinition
from ..factory import ProceduralDataset, register_dataset

DATASET_NAME = "dice"


def compute_probability(dice, target):
    """
    Computes the probability of rolling a total of at least `target`
    when rolling dice specified in the list `dice`. Each element in dice
    is the number of sides on that die. The computation is done via dynamic programming.
    Returns the probability as a fraction (numerator, denominator) and as a float.
    """
    # dp[i][s] = number of ways to get sum s using the first i dice.
    # We use only one dictionary for the current dp state.
    dp = {0: 1}
    for sides in dice:
        new_dp = {}
        for current_sum, count in dp.items():
            # Each die gives a number from 1 to sides.
            for face in range(1, sides + 1):
                new_sum = current_sum + face
                new_dp[new_sum] = new_dp.get(new_sum, 0) + count
        dp = new_dp

    total_outcomes = reduce(lambda a, b: a * b, dice, 1)
    ways = sum(count for s, count in dp.items() if s >= target)

    # Simplify the fraction (ways / total_outcomes)
    def simplify(n, d):
        common = gcd(n, d)
        return n // common, d // common

    frac = simplify(ways, total_outcomes)
    return frac, ways / total_outcomes


def generate_puzzle(num_dice, max_dice_size, rng):
    """
    Generates a puzzle:
      - It forces one die to have max_dice_size.
      - The other (num_dice-1) dice are chosen randomly between 2 and max_dice_size-1.
      - The dice are then shuffled.
      - The target total is chosen roughly in the middle (but you can adjust the method).

    It then computes the probability of rolling a total at least the target.
    Finally, it prints out the puzzle statement and the answer.
    """

    # Guarantee one die is the maximum.
    dice = [max_dice_size]
    for _ in range(num_dice - 1):
        # Choose a die size randomly from 2 up to max_dice_size-1.
        # (If max_dice_size == 2 then all dice are 2-sided.)
        if max_dice_size > 2:
            die = rng.randint(2, max_dice_size - 1)
        else:
            die = 2
        dice.append(die)

    # Optionally, sort dice in descending order (as is common in puzzles)
    dice.sort(reverse=True)

    # Compute minimum and maximum possible totals.
    min_total = num_dice  # each die gives at least 1
    max_total = sum(dice)

    # Choose a target total. For an interesting puzzle,
    # we choose a target somewhere in the middle third of the range.
    low_target = min_total + (max_total - min_total) // 3
    high_target = min_total + 2 * (max_total - min_total) // 3
    target = rng.randint(low_target, high_target)

    # Compute probability. Don't change this.
    (num, den), probability = compute_probability(dice, target)

    # Create a string representing the dice, e.g., "1d20, 1d17, 1d6" etc.
    dice_str = ", ".join(f"1d{s}" for s in dice)

    # Return the puzzle.
    return {"dice_str": dice_str, "target": target, "num": num, "den": den, "probability": probability}


@dataclass
class DiceConfig:
    """Configuration for dice puzzle generation"""

    num_dice: int = 4
    max_dice_size: int = 20
    seed: Optional[int] = None
    size: int = 500

    def validate(self):
        """Validate configuration parameters"""
        assert self.num_dice >= 1, "num_dice must be gte 1"
        assert self.max_dice_size >= 2, "max_dice_size must be gte 2"


class DiceDataset(ProceduralDataset):
    """Generates Dice-based puzzles with configurable parameters"""

    def __init__(self, config: DiceConfig):
        super().__init__(config=config, seed=config.seed, size=config.size)

    def __getitem__(self, idx: int) -> dict:
        """Generate a single Dice task

        Returns:
            dict with keys:
                - question: str, the task description
                - answer: str, a solution string
                - metadata: dict with generation parameters
        """
        rng = Random(self.seed + idx)
        puzzle = generate_puzzle(self.config.num_dice, self.config.max_dice_size, rng)
        puzzle_str = f"I have these dice: {puzzle['dice_str']}. What are the odds of rolling {puzzle['target']} or higher? (Assume that all dice are rolled at once, and that '1d6' represents one roll of a 6-sided dice.) Please respond with a reduced fraction representing the probability [ex., 1/60]."
        answer_str = f"{puzzle['num']}/{puzzle['den']}"

        return {
            "question": puzzle_str,
            "answer": answer_str,
            "metadata": {
                "source_dataset": DATASET_NAME,
                "source_index": idx,
                "puzzle": puzzle,
                "difficulty": {
                    "num_dice": self.config.num_dice,
                    "max_dice_size": self.config.max_dice_size,
                },
            },
        }

    def score_answer(self, answer: Optional[str], entry: dict[str, Any]) -> float:
        """Determine if the solution provided solves the Dice task.

        The function awards 1.0 for a correct answer.

        Args:
            answer (Optional[str]): The user's answer.
            entry (dict[str, Any]): The original dataset entry containing the correct answer.

        Returns:
            float: The computed score between 0.0 and 1.0.
        """

        if isinstance(answer, str):
            if answer.lower().replace("\n", "") == entry["answer"].lower().replace("\n", ""):
                return 1.0  # Yay

        return 0.0


class DiceCurriculum(BaseCurriculum):
    """Curriculum for dice puzzle generation"""

    def __init__(self):
        super().__init__(DiceCurriculum.__name__, DiceConfig)

        self._define_attributes(
            ScalarAttributeDefinition(
                name="num_dice",
                levels=[4, 6, 8, 10],
                description="Number of dice to roll",
                field_name="num_dice",
            ),
            ScalarAttributeDefinition(
                name="max_dice_size",
                levels=[20, 25, 30, 35],
                description="Maximum number of sides on any die",
                field_name="max_dice_size",
            ),
        )


register_dataset(DATASET_NAME, DiceDataset, DiceConfig, DiceCurriculum)
